# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Route_maps parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible.module_utils.six import iteritems
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _tmplt_route_map_match(config_data):
    if (
        config_data.get("match")
        and not config_data["match"].get("ip")
        and not config_data["match"].get("ipv6")
    ):
        command = []
        match = config_data["match"]
        if match and match.get("additional_paths"):
            cmd = "match additional-paths advertise-set"
            if config_data["match"]["additional_paths"].get("all"):
                cmd += " all"
            if config_data["match"]["additional_paths"].get("best"):
                cmd += " best {best}".format(**config_data["match"]["additional_paths"])
            if config_data["match"]["additional_paths"].get("best_range"):
                cmd += " best-range"
                if config_data["match"]["additional_paths"]["best_range"].get("lower_limit"):
                    cmd += " lower-limit {lower_limit}".format(
                        **config_data["match"]["additional_paths"]["best_range"],
                    )
                if config_data["match"]["additional_paths"]["best_range"].get("upper_limit"):
                    cmd += " upper-limit {upper_limit}".format(
                        **config_data["match"]["additional_paths"]["best_range"],
                    )
            if config_data["match"]["additional_paths"].get("group_best"):
                cmd += " group-best"
            command.append(cmd)
        if match.get("as_path"):
            cmd = "match as-path "
            if match["as_path"].get("acls"):
                temp = []
                for k, v in iteritems(match["as_path"]["acls"]):
                    temp.append(str(v))
                cmd += " ".join(sorted(temp))
            command.append(cmd)
        if match.get("clns"):
            cmd = "match clns"
            if match["clns"].get("address"):
                cmd += " address {address}".format(**match["clns"])
            elif match["clns"].get("next_hop"):
                cmd = " next-hop {next_hop}".format(**match["clns"])
            elif match["clns"].get("route_source"):
                cmd = " route-source {route_source}".format(**match["clns"])
            command.append(cmd)
        if match.get("community"):
            cmd = "match community "
            temp = []
            for k, v in iteritems(match["community"]["name"]):
                temp.append(v)
            cmd += " ".join(sorted(temp))
            if match["community"].get("exact_match"):
                cmd += " exact-match"
            command.append(cmd)
        if match.get("extcommunity"):
            cmd = "match extcommunity "
            temp = []
            for k, v in iteritems(match["extcommunity"]):
                temp.append(v)
            cmd += " ".join(sorted(temp))
            command.append(cmd)
        if match.get("interfaces"):
            cmd = "match interface "
            temp = []
            for k, v in iteritems(match["interfaces"]):
                temp.append(v)
            cmd += " ".join(sorted(temp))
            command.append(cmd)
        if match.get("length"):
            command.append("match length {minimum} {maximum}".format(**match["length"]))
        if match.get("local_preference"):
            cmd = "match local-preference "
            if match["local_preference"].get("value"):
                temp = []
                for k, v in iteritems(match["local_preference"]["value"]):
                    temp.append(v)
                cmd += " ".join(sorted(temp))
            command.append(cmd)
        if match.get("mdt_group"):
            cmd = "match mdt-group "
            if match["mdt_group"].get("acls"):
                temp = []
                for k, v in iteritems(match["mdt_group"]["acls"]):
                    temp.append(v)
                cmd += " ".join(sorted(temp))
            command.append(cmd)
        if match.get("metric"):
            cmd = "match metric"
            if match["metric"].get("external"):
                cmd += " external"
            if match["metric"].get("value"):
                cmd += " {value}".format(**match["metric"])
            if match["metric"].get("deviation"):
                cmd += " +-"
                if match["metric"].get("deviation_value"):
                    cmd += " {deviation_value}".format(**match["metric"])
            command.append(cmd)
        if match.get("mpls_label"):
            command.append("match mpls-label")
        if match.get("policy_lists"):
            cmd = "match policy-list "
            temp = []
            for k, v in iteritems(match["policy_lists"]):
                temp.append(v)
            cmd += " ".join(sorted(temp))
            command.append(cmd)
        if match.get("route_type"):
            cmd = "match route-type"
            if match["route_type"].get("external"):
                cmd += " external"
                if match["route_type"]["external"].get("type_1"):
                    cmd += " type-1"
                elif match["route_type"]["external"].get("type_2"):
                    cmd += " type-2"
            elif match["route_type"].get("internal"):
                cmd += " internal"
            elif match["route_type"].get("level_1"):
                cmd += " level-1"
            elif match["route_type"].get("level_2"):
                cmd += " level-2"
            elif match["route_type"].get("local"):
                cmd += " local"
            elif match["route_type"].get("nssa_external"):
                cmd += " nssa-external"
                if match["route_type"]["nssa_external"].get("type_1"):
                    cmd += " type-1"
                elif match["route_type"]["nssa_external"].get("type_2"):
                    cmd += " type-2"
            command.append(cmd)
        if match.get("rpki"):
            cmd = "match rpki"
            if match["rpki"].get("invalid"):
                cmd += " invalid"
            if match["rpki"].get("not_found"):
                cmd += " not-found"
            if match["rpki"].get("valid"):
                cmd += " valid"
            command.append(cmd)
        if match.get("security_group"):
            cmd = "match security-group"
            if match["security_group"].get("source"):
                cmd += " source tag "
                temp = []
                for k, v in iteritems(match["security_group"]["source"]):
                    temp.append(str(v))
                cmd += " ".join(sorted(temp))
            elif match["security_group"].get("destination"):
                cmd += " destination tag"
                for each in match["destination"]:
                    cmd += " {0}".format(each)
            command.append(cmd)
        if match.get("source_protocol"):
            cmd = "match source-protocol"
            if match["source_protocol"].get("bgp"):
                cmd += " bgp {bgp}".format(**match["source_protocol"])
            if match["source_protocol"].get("connected"):
                cmd += " connected"
            if match["source_protocol"].get("eigrp"):
                cmd += " eigrp {eigrp}".format(**match["source_protocol"])
            if match["source_protocol"].get("isis"):
                cmd += " isis"
            if match["source_protocol"].get("lisp"):
                cmd += " lisp"
            if match["source_protocol"].get("mobile"):
                cmd += " mobile"
            if match["source_protocol"].get("ospf"):
                cmd += " ospf {ospf}".format(**match["source_protocol"])
            if match["source_protocol"].get("ospfv3"):
                cmd += " ospfv3 {ospfv3}".format(**match["source_protocol"])
            if match["source_protocol"].get("rip"):
                cmd += " rip"
            if match["source_protocol"].get("static"):
                cmd += " static"
            command.append(cmd)
        if match.get("tag"):
            cmd = "match tag"
            if match["tag"].get("tag_list"):
                cmd += " list"
                for each in match["tag"]["tag_list"]:
                    cmd += " {0}".format(each)
            elif match["tag"].get("value"):
                for each in match["tag"]["value"]:
                    cmd += " {0}".format(each)
            command.append(cmd)
        if match.get("track"):
            command.append("match track {track}".format(**match))
        return command


def _tmplt_route_map_match_ip(config_data):
    if config_data.get("match") and config_data["match"].get("ip"):

        def construct_cmd_from_list(cmd, config):
            temp = []
            for k, v in iteritems(config):
                temp.append(v)
            cmd += " " + " ".join(sorted(temp))
            return cmd

        cmd = "match ip"
        if config_data["match"]["ip"].get("address"):
            cmd += " address"
            if config_data["match"]["ip"]["address"].get("prefix_lists"):
                cmd += " prefix-list"
                cmd = construct_cmd_from_list(
                    cmd,
                    config_data["match"]["ip"]["address"]["prefix_lists"],
                )
            elif config_data["match"]["ip"]["address"].get("acls"):
                cmd = construct_cmd_from_list(cmd, config_data["match"]["ip"]["address"]["acls"])
        if config_data["match"]["ip"].get("flowspec"):
            cmd += " flowspec"
            if config_data["match"]["ip"]["flowspec"].get("dest_pfx"):
                cmd += " dest-pfx"
            elif config_data["match"]["ip"]["flowspec"].get("src_pfx"):
                cmd += " src-pfx"
            if config_data["match"]["ip"]["flowspec"].get("prefix_lists"):
                cmd += " prefix-list"
                cmd = construct_cmd_from_list(
                    cmd,
                    config_data["match"]["ip"]["flowspec"]["prefix_lists"],
                )
            elif config_data["match"]["ip"]["flowspec"].get("acls"):
                cmd = construct_cmd_from_list(cmd, config_data["match"]["ip"]["flowspec"]["acls"])
        if config_data["match"]["ip"].get("next_hop"):
            cmd += " next-hop"
            if config_data["match"]["ip"]["next_hop"].get("prefix_lists"):
                cmd += " prefix-list"
                cmd = construct_cmd_from_list(
                    cmd,
                    config_data["match"]["ip"]["next_hop"]["prefix_lists"],
                )
            elif config_data["match"]["ip"]["next_hop"].get("acls"):
                cmd = construct_cmd_from_list(cmd, config_data["match"]["ip"]["next_hop"]["acls"])
        if config_data["match"]["ip"].get("redistribution_source"):
            cmd += " redistribution-source"
            if config_data["match"]["ip"]["redistribution_source"].get("prefix_lists"):
                cmd += " prefix-list"
                cmd = construct_cmd_from_list(
                    cmd,
                    config_data["match"]["ip"]["redistribution_source"]["prefix_lists"],
                )
            elif config_data["match"]["ip"]["redistribution_source"].get("acls"):
                cmd = construct_cmd_from_list(
                    cmd,
                    config_data["match"]["ip"]["redistribution_source"]["acls"],
                )
        if config_data["match"]["ip"].get("route_source"):
            cmd += " route-source"
            if config_data["match"]["ip"]["route_source"].get("redistribution_source"):
                cmd += " redistribution-source"
            if config_data["match"]["ip"]["route_source"].get("prefix_lists"):
                cmd += " prefix-list"
                cmd = construct_cmd_from_list(
                    cmd,
                    config_data["match"]["ip"]["route_source"]["prefix_lists"],
                )
            elif config_data["match"]["ip"]["route_source"].get("acls"):
                cmd = construct_cmd_from_list(
                    cmd,
                    config_data["match"]["ip"]["route_source"]["acls"],
                )
        return cmd


def _tmplt_route_map_match_ipv6(config_data):
    if config_data.get("match") and config_data["match"].get("ipv6"):
        cmd = "match ipv6"
        if config_data["match"]["ipv6"].get("address"):
            cmd += " address"
            if config_data["match"]["ipv6"]["address"].get("prefix_list"):
                cmd += " prefix-list {prefix_list}".format(
                    **config_data["match"]["ipv6"]["address"],
                )
            elif config_data["match"]["ipv6"]["address"].get("acl"):
                cmd += " {acl}".format(**config_data["match"]["ipv6"]["address"])
        if config_data["match"]["ipv6"].get("flowspec"):
            cmd += " flowspec"
            if config_data["match"]["ipv6"]["flowspec"].get("dest_pfx"):
                cmd += " dest-pfx"
            elif config_data["match"]["ipv6"]["flowspec"].get("src_pfx"):
                cmd += " src-pfx"
            if config_data["match"]["ipv6"]["flowspec"].get("prefix_list"):
                cmd += " prefix-list {prefix_list}".format(
                    **config_data["match"]["ipv6"]["flowspec"],
                )
            elif config_data["match"]["ipv6"]["flowspec"].get("acl"):
                cmd += " {acl}".format(**config_data["match"]["ipv6"]["flowspec"])
        if config_data["match"]["ipv6"].get("next_hop"):
            cmd += " next-hop"
            if config_data["match"]["ipv6"]["next_hop"].get("prefix_list"):
                cmd += " prefix-list {prefix_list}".format(
                    **config_data["match"]["ipv6"]["next_hop"],
                )
            elif config_data["match"]["ipv6"]["next_hop"].get("acl"):
                cmd += " {acl}".format(**config_data["match"]["ipv6"]["next_hop"])
        if config_data["match"]["ipv6"].get("route_source"):
            cmd += " route-source"
            if config_data["match"]["ipv6"]["route_source"].get("prefix_list"):
                cmd += " prefix-list {prefix_list}".format(
                    **config_data["match"]["ipv6"]["route_source"],
                )
            elif config_data["match"]["ipv6"]["route_source"].get("acl"):
                cmd += " {acl}".format(**config_data["match"]["ipv6"]["route_source"])
        return cmd


def _tmplt_route_map_set(config_data):
    if config_data.get("set"):
        command = []
        set = config_data["set"]
        if set.get("aigp_metric"):
            cmd = "set aigp-metric"
            if set["aigp_metric"].get("value"):
                cmd += " {value}".format(**set["aigp_metric"])
            elif set["aigp_metric"].get("igp_metric"):
                cmd += " igp-metric"
            command.append(cmd)
        if set.get("as_path"):
            cmd = "set as-path"
            if set["as_path"].get("prepend"):
                cmd += " prepend"
                if set["as_path"]["prepend"].get("as_number"):
                    cmd += " {0}".format(set["as_path"]["prepend"].get("as_number"))
                elif set["as_path"]["prepend"].get("last_as"):
                    cmd += " last-as {last_as}".format(**set["as_path"]["prepend"])
            if set["as_path"].get("tag"):
                cmd += " tag"
            command.append(cmd)
        if set.get("automatic_tag"):
            command.append("set automatic-tag")
        if set.get("clns"):
            command.append("set clns next-hop {clns}".format(**set))
        if set.get("comm_list"):
            command.append("set comm-list {comm_list} delete".format(**set))
        if set.get("community"):
            cmd = "set community"
            if set["community"].get("number"):
                cmd += " " + " ".join(i for i in set["community"]["number"])
            if set["community"].get("gshut"):
                cmd += " gshut"
            if set["community"].get("internet"):
                cmd += " internet"
            if set["community"].get("local_as"):
                cmd += " local-as"
            if set["community"].get("no_advertise"):
                cmd += " no-advertise"
            if set["community"].get("no_export"):
                cmd += " no-export"
            if set["community"].get("none"):
                cmd += " none"
            # additive must be set last last
            if set["community"].get("additive"):
                cmd += " additive"
            command.append(cmd)
        if set.get("dampening"):
            command.append(
                "set dampening {penalty_half_time} {reuse_route_val} {suppress_route_val} {max_suppress}".format(
                    **set["dampening"],
                ),
            )
        if set.get("default"):
            command.append("set default interface {default}".format(**set["default"]))
        if set.get("extcomm_list"):
            command.append("set extcomm-list {extcomm_list} delete".format(**set))
        if set.get("extcommunity"):
            if set["extcommunity"].get("cost"):
                cmd = "set extcommunity cost"
                if set["extcommunity"]["cost"].get("igp"):
                    cmd += " igp"
                elif set["extcommunity"]["cost"].get("pre_bestpath"):
                    cmd += " pre-bestpath"
                if set["extcommunity"]["cost"].get("id"):
                    cmd += " {id}".format(**set["extcommunity"]["cost"])
                if set["extcommunity"]["cost"].get("cost_value"):
                    cmd += " {cost_value}".format(**set["extcommunity"]["cost"])
                command.append(cmd)
            if set["extcommunity"].get("rt"):
                cmd = "set extcommunity rt"
                if set["extcommunity"]["rt"].get("range"):
                    cmd += " range {lower_limit} {upper_limit}".format(
                        **set["extcommunity"]["rt"]["range"],
                    )
                elif set["extcommunity"]["rt"].get("address"):
                    cmd += " {address}".format(**set["extcommunity"]["rt"])
                if set["extcommunity"]["rt"].get("additive"):
                    cmd += " additive"
                command.append(cmd)
            if set["extcommunity"].get("soo"):
                command.append("set extcommunity soo {soo}".format(**set["extcommunity"]))
            if set["extcommunity"].get("vpn_distinguisher"):
                cmd = "set extcommunity vpn-distinguisher"
                if set["extcommunity"]["vpn_distinguisher"].get("range"):
                    cmd += " range {lower_limit} {upper_limit}".format(
                        **set["extcommunity"]["vpn_distinguisher"]["range"],
                    )
                elif set["extcommunity"]["vpn_distinguisher"].get("address"):
                    cmd += " {address}".format(**set["extcommunity"]["vpn_distinguisher"])
                if set["extcommunity"]["vpn_distinguisher"].get("additive"):
                    cmd += " additive"
                command.append(cmd)
        if set.get("global"):
            command.append("set global")
        if set.get("interfaces"):
            cmd = "set interface "
            temp = []
            for k, v in iteritems(set["interfaces"]):
                temp.append(v)
            cmd += " ".join(sorted(temp))
            command.append(cmd)
        if set.get("level"):
            cmd = "set level"
            if set["level"].get("level_1"):
                cmd += " level-1"
            elif set["level"].get("level_1_2"):
                cmd += " level-1-2"
            elif set["level"].get("level_2"):
                cmd += " level-2"
            elif set["level"].get("nssa_only"):
                cmd += " nssa-only"
        if set.get("lisp"):
            command.append("set lisp locator-set {lisp}".format(**set))
        if set.get("local_preference"):
            command.append("set local-preference {local_preference}".format(**set))
        if set.get("metric"):
            cmd = "set metric"
            if set["metric"].get("metric_value"):
                cmd += " {metric_value}".format(**set["metric"])
                if set["metric"].get("deviation"):
                    if set["metric"]["deviation"] == "plus":
                        cmd += (
                            " +{eigrp_delay} {metric_reliability} {metric_bandwidth} {mtu}".format(
                                **set["metric"],
                            )
                        )
                    elif set["metric"]["deviation"] == "minus":
                        cmd += (
                            " -{eigrp_delay} {metric_reliability} {metric_bandwidth} {mtu}".format(
                                **set["metric"],
                            )
                        )
            if set["metric"].get("deviation") and not set["metric"].get("eigrp_delay"):
                if set["metric"]["deviation"] == "plus":
                    cmd = "set metric +{metric_value}".format(**set["metric"])
                elif set["metric"]["deviation"] == "minus":
                    cmd = "set metric -{metric_value}".format(**set["metric"])
            command.append(cmd)
        if set.get("metric_type"):
            cmd = "set metric-type"
            if set["metric_type"].get("external"):
                cmd += " external"
            elif set["metric_type"].get("internal"):
                cmd += " internal"
            elif set["metric_type"].get("type_1"):
                cmd += " type-1"
            elif set["metric_type"].get("type_2"):
                cmd += " type-2"
            command.append(cmd)
        if set.get("mpls_label"):
            command.append("set mpls-label")
        if set.get("origin"):
            cmd = "set origin"
            if set["origin"].get("igp"):
                cmd += " igp"
            elif set["origin"].get("incomplete"):
                cmd += " incomplete"
        if set.get("tag"):
            command.append("set tag {tag}".format(**set))
        if set.get("traffic_index"):
            command.append("set traffic-index {traffic_index}".format(**set))
        if set.get("vrf"):
            command.append("set vrf {vrf}".format(**set))
        if set.get("weight"):
            command.append("set weight {weight}".format(**set))
        return command


def _tmplt_route_map_set_ip(config_data):
    if config_data.get("set") and config_data["set"].get("ip"):
        command = []
        set_ip = config_data["set"]["ip"]
        cmd = "set ip"
        if set_ip.get("address"):
            command.append("{0} address prefix-list {address}".format(cmd, **set_ip))
        if set_ip.get("df"):
            command.append("{0} df {df}".format(cmd, **set_ip))
        if set_ip.get("global_route"):
            cmd += " global next-hop"
            if set_ip["global_route"].get("verify_availability"):
                cmd += " verify-availability {address} {sequence} track {track}".format(
                    **set_ip["global_route"]["verify_availability"],
                )
            elif set_ip["global_route"].get("address"):
                cmd += " {address}".format(**set_ip["global_route"])
            command.append(cmd)
        if set_ip.get("next_hop"):
            cmd += " next-hop"
            if set_ip["next_hop"].get("address"):
                command.append("{0} {address}".format(cmd, **set_ip["next_hop"]))
            if set_ip["next_hop"].get("dynamic"):
                command.append("{0} dynamic dhcp".format(cmd))
            if set_ip["next_hop"].get("encapsulate"):
                command.append(
                    "{0} encapsulate l3vpn {encapsulate}".format(cmd, **set_ip["next_hop"]),
                )
            if set_ip["next_hop"].get("peer_address"):
                command.append("{0} peer-address".format(cmd))
            if set_ip["next_hop"].get("recursive"):
                child_cmd = "{0} recursive".format(cmd)
                if set_ip["next_hop"]["recursive"].get("global_route"):
                    child_cmd += " global"
                elif set_ip["next_hop"]["recursive"].get("vrf"):
                    child_cmd += " vrf {vrf}".format(**set_ip["next_hop"]["recursive"])
                if set_ip["next_hop"]["recursive"].get("address"):
                    child_cmd += " {address}".format(**set_ip["next_hop"]["recursive"])
                command.append(child_cmd)
            if set_ip["next_hop"].get("self"):
                command.append("{0} self".format(cmd))
            if set_ip["next_hop"].get("verify_availability"):
                command.append(
                    "{0} verify-availability {address} {sequence} track {track}".format(
                        cmd,
                        **set_ip["next_hop"]["verify_availability"],
                    ),
                )
        if set_ip.get("precedence"):
            cmd += " precedence"
            if set_ip["precedence"].get("critical"):
                cmd += " critical"
            elif set_ip["precedence"].get("flash"):
                cmd += " flash"
            elif set_ip["precedence"].get("flash_override"):
                cmd += " flash-override"
            elif set_ip["precedence"].get("immediate"):
                cmd += " immediate"
            elif set_ip["precedence"].get("internet"):
                cmd += " internet"
            elif set_ip["precedence"].get("network"):
                cmd += " network"
            elif set_ip["precedence"].get("priority"):
                cmd += " priority"
            elif set_ip["precedence"].get("routine"):
                cmd += " routine"
            command.append(cmd)
        if set_ip.get("qos_group"):
            command.append("{0} qos-group {qos_group}".format(cmd, **set_ip))
        if set_ip.get("tos"):
            cmd += " tos"
            if set_ip["tos"].get("max_reliability"):
                cmd += " max-reliability"
            elif set_ip["tos"].get("max_throughput"):
                cmd += " max-throughput"
            elif set_ip["tos"].get("min_delay"):
                cmd += " min-delay"
            elif set_ip["tos"].get("min_monetary_cost"):
                cmd += " min-monetary-cost"
            elif set_ip["tos"].get("normal"):
                cmd += " normal"
            command.append(cmd)
        if set_ip.get("vrf"):
            cmd += " vrf {vrf} next-hop".format(**set_ip)
            if set_ip["vrf"].get("verify_availability").get("address"):
                cmd += " verify-availability {address} {sequence} track {track}".format(
                    **set_ip["vrf"]["verify_availability"],
                )
            elif set_ip["vrf"].get("address"):
                cmd += " {address}".format(**set_ip["vrf"])
            command.append(cmd)
        return command


def _tmplt_route_map_set_ipv6(config_data):
    if config_data.get("set") and config_data["set"].get("ipv6"):
        set_ipv6 = config_data["set"]["ipv6"]
        cmd = "set ipv6"
        if set_ipv6.get("address"):
            cmd += " address prefix-list {address}".format(**set_ipv6)
        if set_ipv6.get("default"):
            cmd += " default"
        if set_ipv6.get("global_route"):
            cmd += " global next-hop"
            if set_ipv6["global_route"].get("verify_availability"):
                cmd += " verify-availability {address} {sequence} track {track}".format(
                    **set_ipv6["global_route"]["verify_availability"],
                )
            elif set_ipv6["global_route"].get("address"):
                cmd += " {address}".format(**set_ipv6["global_route"])
        if set_ipv6.get("next_hop"):
            cmd += " next-hop"
            if set_ipv6["next_hop"].get("address"):
                cmd += " {address}".format(**set_ipv6["next_hop"])
            if set_ipv6["next_hop"].get("encapsulate"):
                cmd += " encapsulate l3vpn {encapsulate}".format(**set_ipv6["next_hop"])
            if set_ipv6["next_hop"].get("peer_address"):
                cmd += " peer-address"
        if set_ipv6.get("precedence"):
            cmd += " precedence {precedence}".format(**set_ipv6)
        if set_ipv6.get("vrf"):
            cmd += (
                " vrf {vrf} next-hop verify-availability {address} {sequence} track {track}".format(
                    **set_ipv6["vrf"]["verify_availability"],
                )
            )
        return cmd


class Route_mapsTemplate(NetworkTemplate):
    def __init__(self, lines=None):
        super(Route_mapsTemplate, self).__init__(lines=lines, tmplt=self)

    PARSERS = [
        {
            "name": "route_map",
            "getval": re.compile(
                r"""
                ^route-map*
                \s*(?P<route_map>\S+)*
                \s*(?P<action>deny|permit)*
                \s*(?P<sequence>\d+)*
                (\s|$)""",
                re.VERBOSE,
            ),
            "setval": "",
            "result": {
                "{{ route_map }}": {
                    "route_map": "{{ route_map }}",
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {"action": "{{ action }}", "sequence": "{{ sequence }}"},
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "continue_entry",
            "getval": re.compile(
                r"""
                \s+continue*
                \s*(?P<entry_sequence>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": "continue {{ continue_entry.entry_sequence }}",
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {
                            "continue_entry": {
                                "set": "{{ True if entry_sequence is not defined }}",
                                "entry_sequence": "{{ entry_sequence }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "description",
            "getval": re.compile(
                r"""
                \s+description*
                \s*(?P<description>\S.*)*
                $""",
                re.VERBOSE,
            ),
            "setval": "description {{ description }}",
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {"description": "{{ description }}"},
                    },
                },
            },
        },
        {
            "name": "match",
            "getval": re.compile(
                r"""
                \s+match*
                \s*(?P<additional_paths>additional-paths\sadvertise-set\s\S.*)*
                \s*(?P<as_path>as-path.*|as-path)*
                \s*(?P<clns>clns\s(address\s\S+|next-hop\s\S+|route-source\s\S+))*
                \s*(?P<community>community\s\S.*)*
                \s*(?P<extcommunity>extcommunity\s\S.*)*
                \s*(?P<interfaces>interface\s\S.*)*
                \s*(?P<length>length\s\d+\s\d+)*
                \s*(?P<local_preference>local-preference\s\d.*|local-preference)*
                \s*(?P<mdt_group>mdt-group\s\S.*|mdt-group)*
                \s*(?P<metric>metric\sexternal\s\S.*|metric\s\d+\S.*)*
                \s*(?P<mpls_label>mpls-label)*
                \s*(?P<policy_list>policy-list\s\S.*)*
                \s*(?P<route_type>route-type\s(external\s(type-1|type-2)|internal|level-1|level-2|local|nssa-external\s(type-1|type-2)))*
                \s*(?P<rpki>rpki\s(invalid|not-found|valid))*
                \s*(?P<security_group>security-group\s(destination\stag\s\d.*|source\stag\s\d.*))*
                \s*(?P<source_protocol>source-protocol\s\S.*)*
                \s*(?P<tag>tag\slist\s\S.*|tag\s\S.*)*
                \s*(?P<track>track\s*\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_route_map_match,
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {
                            "match": {
                                "additional_paths": {
                                    "all": "{{ True if additional_paths is defined and 'all' in additional_paths }}",
                                    "best": "{{ additional_paths.split('best ')[1].split(' ')[0]|int if additional_paths is defined and\
                                             'best' in additional_paths and 'best-range' not in additional_paths }}",
                                    "best_range": {
                                        "lower_limit": "{{ additional_paths.split('best-range ')[1].split(' ')[0]|int if additional_paths is defined and\
                                                 'best-range' in additional_paths }}",
                                        "upper_limit": "{{ additional_paths.split('best-range ')[1].split(' ')[1]|int if additional_paths is defined and\
                                                 'best-range' in additional_paths }}",
                                    },
                                    "group_best": "{{ True if additional_paths is defined and 'group-best' in additional_paths }}",
                                },
                                "as_path": {
                                    "set": "{{ True if as_path is defined and as_path.split(' ')|length == 1 }}",
                                    "acls": "{{ as_path.split('as-path ')[1].split(' ') if as_path is defined and as_path.split(' ')|length > 1 }}",
                                },
                                "clns": {
                                    "address": "{{ clns.split('clns address ')[1] if clns is defined }}",
                                    "next_hop": "{{ clns.split('clns next-hop ')[1] if clns is defined }}",
                                    "route_source": "{{ clns.split('clns route-source ')[1] if clns is defined }}",
                                },
                                "community": {
                                    "name": "{{ community.split('community ')[1].split(' exact-match')[0].split(' ') if community is defined }}",
                                    "exact_match": "{{ True if community is defined and 'exact-match' in community }}",
                                },
                                "extcommunity": "{{ extcommunity.split('extcommunity ')[1].split(' ') if extcommunity is defined }}",
                                "interfaces": "{{ interfaces.split('interface ')[1].split(' ') if interfaces is defined }}",
                                "length": {
                                    "minimum": "{{ length.split(' ')[1] if length is defined }}",
                                    "maximum": "{{ length.split(' ')[2] if length is defined }}",
                                },
                                "local_preference": {
                                    "set": "{{ True if local_preference is defined and local_preference.split(' ')|length == 1 }}",
                                    "value": "{{ local_preference.split('local-preference ')[1].split(' ') if local_preference is defined }}",
                                },
                                "mdt_group": {
                                    "set": "{{ True if mdt_group is defined and mdt_group.split(' ')|length == 1 }}",
                                    "acls": "{{ mdt_group.split('mdt-group ')[1].split(' ') if mdt_group is defined }}",
                                },
                                "metric": {
                                    "external": "{{ True if metric is defined and 'external' in metric.split(' ') }}",
                                    "value": "{% if metric is defined and 'external' not in metric.split(' ') %}{{ metric.split(' ')[1] }}\
                                            {% elif metric is defined and 'external' in metric.split(' ') %}{{ metric.split(' ')[2] }}\
                                            {% endif %}",
                                    "deviation": "{{ True if metric is defined and '+-' in metric }}",
                                    "deviation_value": "{% if metric is defined and 'external' in metric and '+-' in metric %}{{ metric.split(' ')[4] }}\
                                            {% elif metric is defined and 'external' not in metric and '+-' in metric %}{{ metric.split(' ')[3] }}{% endif %}",
                                },
                                "mpls_label": "{{ True if mpls_label is defined }}",
                                "policy_lists": "{{ policy_list.split('policy-list ')[1].split(' ') if policy_list is defined }}",
                                "route_type": {
                                    "external": {
                                        "set": "{{ True if route_type is defined and 'type-1' not in route_type and 'type-2' not in route_type}}",
                                        "type_1": "{{ True if route_type is defined and 'type-1' in route_type }}",
                                        "type_2": "{{ True if route_type is defined and 'type-2' in route_type }}",
                                    },
                                    "internal": "{{ True if route_type is defined and 'internal' in route_type }}",
                                    "level_1": "{{ True if route_type is defined and 'level-1' in route_type }}",
                                    "level_2": "{{ True if route_type is defined and 'level-2' in route_type }}",
                                    "local": "{{ True if route_type is defined and 'local' in route_type }}",
                                    "nssa_external": {
                                        "set": "{{ True if route_type is defined and 'type-1' not in route_type and 'type-2' not in route_type}}",
                                        "type_1": "{{ True if route_type is defined and 'type-1' in route_type }}",
                                        "type_2": "{{ True if route_type is defined and 'type-2' in route_type }}",
                                    },
                                },
                                "rpki": {
                                    "invalid": "{{ True if rpki is defined and 'invalid' in rpki and 'valid' not in rpki.split(' ') }}",
                                    "not_found": "{{ True if rpki is defined and 'not-found' in rpki }}",
                                    "valid": "{{ True if rpki is defined and 'valid' in rpki and 'invalid' not in rpki.split(' ') }}",
                                },
                                "security_group": {
                                    "destination": "{{ security_group.split('destination tag ')[1].split(' ') if security_group is defined and\
                                        'destination' in security_group }}",
                                    "source": "{{ security_group.split('source tag ')[1].split(' ') if security_group is defined and\
                                        'source' in security_group }}",
                                },
                                "source_protocol": {
                                    "bgp": "{{ source_protocol.split('bgp ')[1].split(' ')[0] if source_protocol is defined and 'bgp' in source_protocol }}",
                                    "connected": "{{ True if source_protocol is defined and 'connected' in source_protocol }}",
                                    "eigrp": "{{ source_protocol.split('eigrp ')[1].split(' ')[0] if source_protocol is defined and\
                                            'eigrp' in source_protocol }}",
                                    "isis": "{{ True if source_protocol is defined and 'isis' in source_protocol }}",
                                    "lisp": "{{ True if source_protocol is defined and 'lisp' in source_protocol }}",
                                    "mobile": "{{ True if source_protocol is defined and 'mobile' in source_protocol }}",
                                    "ospf": "{{ source_protocol.split('ospf ')[1].split(' ')[0] if source_protocol is defined and\
                                            'ospf' in source_protocol }}",
                                    "ospfv3": "{{ source_protocol.split('ospfv3 ')[1].split(' ')[0] if source_protocol is defined and\
                                            'ospfv3' in source_protocol }}",
                                    "rip": "{{ True if source_protocol is defined and 'rip' in source_protocol }}",
                                    "static": "{{ True if source_protocol is defined and 'static' in source_protocol }}",
                                },
                                "tag": {
                                    "value": "{{ tag.split('tag ')[1].split(' ') if tag is defined and 'list' not in tag }}",
                                    "tag_list": "{{ tag.split('tag list ')[1].split(' ') if tag is defined and 'list' in tag }}",
                                },
                                "track": "{{ track.split('track ')[1] if track is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "match.ip",
            "getval": re.compile(
                r"""
                \s+match*
                \s*(?P<ip>ip)*
                \s*(?P<address>address\sprefix-list\s\S.*|address\s\S.*)*
                \s*(?P<flowspec>flowspec\sdest-pfx\s(prefix-list\s\S.*|\S.*)|flowspec\ssrc-pfx\s(prefix-list\s\S.*|\S.*))*
                \s*(?P<next_hop>next-hop\sprefix-list\s\S.*|next-hop\s\S.*|next-hop)*
                \s*(?P<redistribution_source>redistribution-source\sprefix-list\s\S.*|redistribution-source\s\S.*|redistribution-source)*
                \s*(?P<route_source>route-source\sredistribution-source\sprefix-list\s\S.*|route-source\sredistribution-source\s\S.*|route-source\sprefix-list\s\S.*|route-source\s\S.*|route-source\sredistribution-source|route-source)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_route_map_match_ip,
            "compval": "match",
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {
                            "match": {
                                "ip": {
                                    "address": {
                                        "acls": "{{ address.split('address ')[1].split(' ') if address is defined and\
                                            'prefix-list' not in address else none }}",
                                        "prefix_lists": "{{ address.split('address prefix-list ')[1].split(' ') if address is defined and\
                                            'prefix-list' in address else None }}",
                                    },
                                    "flowspec": {
                                        "dest_pfx": "{{ True if flowspec is defined and 'dest-pfx' in flowspec }}",
                                        "src_pfx": "{{ True if flowspec is defined and 'src-pfx' in flowspec }}",
                                        "acls": "{{ flowspec.split('flowspec ')[1].split(' ')|d() if flowspec is defined and\
                                            'prefix-list' not in flowspec else '' }}",
                                        "prefix_lists": "{{ flowspec.split('flowspec prefix-list ')[1].split(' ')|d() if flowspec is defined and\
                                            'prefix-list' in flowspec else ''}}",
                                    },
                                    "next_hop": {
                                        "set": "{{ True if next_hop is defined and next_hop.split(' ')|length == 1 }}",
                                        "acls": "{{ next_hop.split('next-hop ')[1].split(' ') if next_hop is defined and\
                                            'prefix-list' not in next_hop else '' }}",
                                        "prefix_lists": "{{ next_hop.split('next-hop prefix-list ')[1].split(' ') if next_hop is defined and\
                                             'prefix-list' in next_hop and next_hop.split('next-hop prefix-list ')[1] is not none else '' }}",
                                    },
                                    "redistribution_source": {
                                        "set": "{{ True if redistribution_source is defined and redistribution_source.split(' ')|length == 1 }}",
                                        "acls": "{{ redistribution_source.split('redistribution-source ')[1].split(' ')|d()\
                                            if redistribution_source is defined and 'prefix-list' not in redistribution_source else '' }}",
                                        "prefix_lists": "{{ redistribution_source.split('redistribution-source prefix-list ')[1].split(' ')|d()\
                                            if redistribution_source is defined and 'prefix-list' in redistribution_source else '' }}",
                                    },
                                    "route_source": {
                                        "set": "{{ True if route_source is defined and route_source.split(' ')|length == 1 }}",
                                        "redistribution_source": "{{ True if route_source is defined and 'redistribution-source' in route_source }}",
                                        "acls": "{{ route_source.split('route-source ')[1].split(' ') if route_source is defined and\
                                                'prefix-list' not in route_source else '' }}",
                                        "prefix_lists": "{{ route_source.split('route-source prefix-list ')[1].split(' ') if route_source is defined and\
                                                'prefix-list' in route_source else '' }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "match.ipv6",
            "getval": re.compile(
                r"""
                \s+match*
                \s*(?P<ipv6>ipv6)*
                \s*(?P<address>address\sprefix-list\s\S.*|address\s\S.*)*
                \s*(?P<flowspec>flowspec\sdest-pfx\s(prefix-list\s\S.*|\S.*)|flowspec\ssrc-pfx\s(prefix-list\s\S.*|\S.*))*
                \s*(?P<next_hop>next-hop\sprefix-list\s\S.*|next-hop\s\S.*)*
                \s*(?P<route_source>route-source\sprefix-list\s\S.*|route-source\s\S.*)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_route_map_match_ipv6,
            "compval": "match",
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {
                            "match": {
                                "ipv6": {
                                    "address": {
                                        "acl": "{{ address.split('address ')[1] if address is defined and 'prefix-list' not in address }}",
                                        "prefix_list": "{{ address.split('address prefix-list ')[1] if address is defined and 'prefix-list' in address }}",
                                    },
                                    "flowspec": {
                                        "dest_pfx": "{{ True if flowspec is defined and 'dest-pfx' in flowspec }}",
                                        "src_pfx": "{{ True if flowspec is defined and 'src-pfx' in flowspec }}",
                                        "acl": "{{ flowspec.split('flowspec ')[1] if flowspec is defined and 'prefix-list' not in flowspec }}",
                                        "prefix_list": "{{ flowspec.split('flowspec prefix-list ')[1] if flowspec is defined and 'prefix-list' in flowspec }}",
                                    },
                                    "next_hop": {
                                        "acl": "{{ next_hop.split('next-hop ')[1] if next_hop is defined and 'prefix-list' not in next_hop }}",
                                        "prefix_list": "{{ next_hop.split('next-hop prefix-list ')[1] if next_hop is defined and 'prefix-list' in next_hop }}",
                                    },
                                    "route_source": {
                                        "acl": "{{ route_source.split('route-source ')[1] if route_source is defined and 'prefix-list' not in route_source }}",
                                        "prefix_list": "{{ route_source.split('route-source prefix-list ')[1] if route_source is defined and\
                                                'prefix-list' in route_source }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "set",
            "getval": re.compile(
                r"""
                \s+set*
                \s*(?P<aigp_metric>aigp-metric\sigp-metric|aigp-metric\s\d+)*
                \s*(?P<as_path>as-path\s(prepend\s(last-as\s\d+|\d+(?:\s\d+)*)|tag))*
                \s*(?P<automatic_tag>automatic-tag)*
                \s*(?P<clns>clns\snext-hop\s\S.*)*
                \s*(?P<comm_list>comm-list\s\S+\sdelete)*
                \s*(?P<community>community\s\S.*)*
                \s*(?P<dampening>dampening\s\d+\s\d+\s\d+\s\d+)*
                \s*(?P<default>default\sinterface\s\S.*)*
                \s*(?P<extcomm_list>extcomm-list\s\S+\sdelete)*
                \s*(?P<extcommunity>extcommunity\s\S.*)*
                \s*(?P<global>global)*
                \s*(?P<interfaces>interface\s\S.*)*
                \s*(?P<level>level\s(level-1-2|level-1|level-2|nssa-only))*
                \s*(?P<lisp>lisp\slocator-set\s\S+)*
                \s*(?P<local_preference>local-preference\s\d+)*
                \s*(?P<metric>metric\s\S.*)*
                \s*(?P<metric_type>metric-type\s(external|internal|type-1|type-2))*
                \s*(?P<mpls_label>mpls-label)*
                \s*(?P<origin>origin\s(igp|incomplete))*
                \s*(?P<tag>tag\s(([0-9]{1,3}\.?){4}|\d+))*
                \s*(?P<traffic_index>traffic-index\s\d+)*
                \s*(?P<vrf>vrf\s\S+)*
                \s*(?P<weight>weight\s\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_route_map_set,
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {
                            "set": {
                                "aigp_metric": {
                                    "value": "{{ aigp_metric.split('aigp-metric ')[1] if aigp_metric is defined and\
                                            'igp-metric' not in aigp_metric.split(' ') }}",
                                    "igp_metric": "{{ True if aigp_metric is defined and 'igp-metric' in aigp_metric.split(' ') }}",
                                },
                                "as_path": {
                                    "prepend": {
                                        "as_number": "{{ as_path.split('as-path prepend ')[1].split(' ')\
                                            if as_path is defined and 'prepend' in as_path and 'last-as' not in as_path }}",
                                        "last_as": "{{ as_path.split('as-path prepend last-as ')[1] if as_path is defined and 'prepend' in as_path and\
                                                'last-as' in as_path }}",
                                    },
                                    "tag": "{{ True if as_path is defined and 'tag' in as_path }}",
                                },
                                "automatic_tag": "{{ True if automatic_tag is defined }}",
                                "clns": "{{ clns.split('clns next-hop ')[1] if clns is defined }}",
                                "comm_list": "{{ comm_list.split(' ')[1] if comm_list is defined }}",
                                "community": {
                                    "number": "{{ community.split(' ')[1:]|reject('in',\
                                        ['additive','gshut','internet','local-AS','no-advertise','no-export','none']\
                                    )|join(' ')}}",
                                    "additive": "{{ True if community is defined and 'additive' in community }}",
                                    "gshut": "{{ True if community is defined and 'gshut' in community }}",
                                    "internet": "{{ True if community is defined and 'internet' in community }}",
                                    "local_as": "{{ True if community is defined and 'local-AS' in community }}",
                                    "no_advertise": "{{ True if community is defined and 'no-advertise' in community }}",
                                    "no_export": "{{ True if community is defined and 'no-export' in community }}",
                                    "none": "{{ True if community is defined and 'none' in community }}",
                                },
                                "dampening": {
                                    "penalty_half_time": "{{ dampening.split(' ')[1] if dampening is defined }}",
                                    "reuse_route_val": "{{ dampening.split(' ')[2] if dampening is defined }}",
                                    "suppress_route_val": "{{ dampening.split(' ')[3] if dampening is defined }}",
                                    "max_suppress": "{{ dampening.split(' ')[4] if dampening is defined }}",
                                },
                                "default": "{{ default.split('default interface ')[1] if default is defined }}",
                                "extcomm_list": "{{ extcomm_list.split(' ')[1] if extcomm_list is defined }}",
                                "extcommunity": {
                                    "cost": {
                                        "id": "{%- if extcommunity is defined and 'cost' in extcommunity and\
                                            'igp' not in extcommunity and 'pre-bestpath' not in extcommunity -%} {{ extcommunity.split(' ')[2] }}\
                                                {%- elif extcommunity is defined and 'cost' in extcommunity and ('igp' in extcommunity or\
                                                    'pre-bestpath' in extcommunity) -%} {{ extcommunity.split(' ')[3] }} {%- endif -%}",
                                        "cost_value": "{% if extcommunity is defined and 'cost' in extcommunity and 'igp' not in extcommunity and\
                                            'pre-bestpath' not in extcommunity %} {{ extcommunity.split(' ')[3] }}\
                                                {% elif extcommunity is defined and 'cost' in extcommunity and ('igp' in extcommunity or\
                                                    'pre-bestpath' in extcommunity) %} {{ extcommunity.split(' ')[4] }} {% endif %}",
                                        "igp": "{{ True if extcommunity is defined and 'cost' in extcommunity and 'igp' in extcommunity }}",
                                        "pre_bestpath": "{{ True if extcommunity is defined and 'cost' in extcommunity and 'pre-bestpath' in extcommunity }}",
                                    },
                                    "rt": {
                                        "address": "{{ extcommunity.split(' ')[2] if extcommunity is defined and 'rt' in extcommunity and\
                                                'range' not in extcommunity }}",
                                        "range": {
                                            "lower_limit": "{{ extcommunity.split('range ')[1].split(' ')[0] if extcommunity is defined and\
                                                    'rt' in extcommunity and 'range' in extcommunity }}",
                                            "upper_limit": "{{ extcommunity.split('range ')[1].split(' ')[1] if extcommunity is defined and\
                                                    'rt' in extcommunity and 'range' in extcommunity }}",
                                        },
                                        "additive": "{{ True if extcommunity is defined and 'rt' in extcommunity and 'additive' in extcommunity  }}",
                                    },
                                    "soo": "{{ extcommunity.split(' ')[2] if extcommunity is defined and 'soo' in extcommunity }}",
                                    "vpn_distinguisher": {
                                        "address": "{{ extcommunity.split(' ')[2] if extcommunity is defined and\
                                                'vpn-distinguisher' in extcommunity and 'range' not in extcommunity }}",
                                        "range": {
                                            "lower_limit": "{{ extcommunity.split('range ')[1].split(' ')[0] if extcommunity is defined and\
                                                    'vpn-distinguisher' in extcommunity and 'range' in extcommunity }}",
                                            "upper_limit": "{{ extcommunity.split('range ')[1].split(' ')[1] if extcommunity is defined and\
                                                    'vpn-distinguisher' in extcommunity and 'range' in extcommunity }}",
                                        },
                                        "additive": "{{ True if extcommunity is defined and 'vpn-distinguisher' in extcommunity and\
                                                 'additive' in extcommunity }}",
                                    },
                                },
                                "global_route": "{{ True if global is defined }}",
                                "interfaces": "{{ interfaces.split('interface ')[1].split(' ') if interfaces is defined }}",
                                "level": {
                                    "level_1": "{{ True if level is defined and 'level-1' in level and 'level-1-2' not in level }}",
                                    "level_1_2": "{{ True if level is defined and 'level-1-2' in level }}",
                                    "level_2": "{{ True if level is defined and 'level-2' in level }}",
                                    "nssa_only": "{{ True if level is defined and 'nssa-only' in level }}",
                                },
                                "lisp": "{{ lisp.split('lisp locator-set ')[1] if lisp is defined }}",
                                "local_preference": "{{ local_preference.split('local-preference ')[1] if local_preference is defined }}",
                                "metric": {
                                    "deviation": "{%- if metric is defined and '+' in metric -%}{{ 'plus' }}\
                                                {%- elif metric is defined and '-' in metric -%}{{ 'minus' }}{%- endif -%}",
                                    "metric_value": "{{ metric.split(' ')[1] if metric is defined and\
                                             (metric.split(' ')[1] != '+' or metric.split(' ')[1] != '-') }}",
                                    "eigrp_delay": "{% if metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                            {{ metric.split('+')[1].split(' ')[0] }}\
                                                {% elif metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                                    {{ metric.split('-')[1].split(' ')[0] }}\
                                                    {% endif %}",
                                    "metric_reliability": "{% if metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                            {{ metric.split('+')[1].split(' ')[1] }}\
                                                {% elif metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                                    {{ metric.split('-')[1].split(' ')[1] }}\
                                                    {% endif %}",
                                    "metric_bandwidth": "{% if metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                            {{ metric.split('+')[1].split(' ')[2] }}\
                                                {% elif metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                                    {{ metric.split('-')[1].split(' ')[2] }}\
                                                    {% endif %}",
                                    "mtu": "{% if metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                            {{ metric.split('+')[1].split(' ')[3] }}\
                                                {% elif metric is defined and metric.split(' ')|length > 2 and '+' in metric %}\
                                                    {{ metric.split('-')[1].split(' ')[3] }}\
                                                    {% endif %}",
                                },
                                "metric_type": {
                                    "external": "{{ True if metric_type is defined and 'external' in metric_type }}",
                                    "internal": "{{ True if metric_type is defined and 'internal' in metric_type }}",
                                    "type_1": "{{ True if metric_type is defined and 'type-1' in metric_type }}",
                                    "type_2": "{{ True if metric_type is defined and 'type-2' in metric_type }}",
                                },
                                "mpls_label": "{{ True if mpls_label is defined }}",
                                "origin": {
                                    "igp": "{{ True if origin is defined and 'igp' in origin }}",
                                    "incomplete": "{{ True if origin is defined and 'incomplete' in origin }}",
                                },
                                "tag": "{{ tag.split('tag ')[1] if tag is defined }}",
                                "traffic_index": "{{ traffic_index.split('traffic-index ')[1] if traffic_index is defined }}",
                                "vrf": "{{ vrf.split('vrf ')[1] if vrf is defined }}",
                                "weight": "{{ weight.split('weight ')[1] if weight is defined }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "set.ip",
            "getval": re.compile(
                r"""
                \s+set*
                \s*(?P<ip>ip)*
                \s*(?P<address>address\sprefix-list\s\S+)*
                \s*(?P<default>default)*
                \s*(?P<df>df\s\d)*
                \s*(?P<global>global\snext-hop\s(verify-availability\s([0-9]{1,3}\.?){4}\s\d+\strack\s\d+|(([0-9]{1,3}\.?){4}).*))*
                \s*(?P<precedence>precedence\s(critical|flash|flash-override|immediate|internet|network|priority|routine)|precedence)*
                \s*(?P<qos_group>qos_group\s\d+)*
                \s*(?P<tos>tos\s(max-reliability|max-throughput|min-delay|min-monetary-cost|normal)|tos)*
                \s*(?P<vrf>vrf\s\S+\snext-hop\s\S.*)*
                \s*(?P<next_hop>next-hop\s\S.*)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_route_map_set_ip,
            "compval": "set",
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {
                            "set": {
                                "ip": {
                                    "address": "{{ address.split('address prefix-list ')[1] if address is defined }}",
                                    "default": "{{ True if default is defined }}",
                                    "df": "{{ df.split('df ')[1] if df is defined }}",
                                    "global_route": {
                                        "address": "{% if global is defined and 'verify-availability' not in global %}{{ global.split('global next-hop ')[1] }}\
                                                        {% elif global is defined and 'verify-availability' in global %}{{ global.split(' ')[3] }}{% endif %}",
                                        "verify_availability": {
                                            "address": "{{ global.split(' ')[3] if global is defined and 'verify-availability' in global }}",
                                            "sequence": "{{ global.split(' ')[4] if global is defined and 'verify-availability' in global }}",
                                            "track": "{{ global.split('track ')[1] if global is defined and 'verify-availability' in global }}",
                                        },
                                    },
                                    "next_hop": {
                                        "address": "{{ next_hop.split('next-hop ')[1] if next_hop is defined and 'peer-address' not in next_hop and\
                                                 'self' not in next_hop and next_hop.split(' ')|length == 2 }}",
                                        "dynamic": "{{ True if next_hop is defined and 'dynamic dhcp' in next_hop }}",
                                        "encapsulate": "{{ next_hop.split('next-hop encapsulate l3vpn ')[1] if next_hop is defined and\
                                                'encapsulate' in next_hop }}",
                                        "peer_address": "{{ True if next_hop is defined and 'peer-address' in next_hop }}",
                                        "recursive": {
                                            "global_route": "{{ True if next_hop is defined and 'global' in next_hop.split(' ') }}",
                                            "vrf": "{{ next_hop.split(' ')[3] if next_hop is defined and 'vrf' in next_hop }}",
                                            "address": "{%- if next_hop is defined and 'global' in next_hop.split(' ') -%}\
                                                {{  next_hop.split('next-hop recursive global ')[1] }}\
                                                {%- elif next_hop is defined and 'vrf' in next_hop.split(' ') -%}{{  next_hop.split(' ')[4] }}\
                                                {%- elif next_hop is defined and 'vrf' not in next_hop.split(' ') and 'global' not in next_hop.split(' ') -%}\
                                                    {{ next_hop.split(' ')[2] }} {%- endif -%}",
                                        },
                                        "self": "{{ True if next_hop is defined and 'self' in next_hop }}",
                                        "verify_availability": {
                                            "set": "{{ True if next_hop is defined and 'verify-availability' in next_hop and 'track' not in next_hop }}",
                                            "address": "{{ next_hop.split(' ')[2] if next_hop is defined and 'verify-availability' in next_hop and\
                                                     'track' in next_hop }}",
                                            "sequence": "{{ next_hop.split(' ')[3] if next_hop is defined and 'verify-availability' in next_hop and\
                                                     'track' in next_hop }}",
                                            "track": "{{ next_hop.split('track ')[1] if next_hop is defined and 'verify-availability' in next_hop and\
                                                     'track' in next_hop }}",
                                        },
                                    },
                                    "precedence": {
                                        "set": "{{ True if precedence is defined and precedence.split(' ')|length == 1 }}",
                                        "critical": "{{ True if precedence is defined and 'critical' in precedence }}",
                                        "flash": "{{ True if precedence is defined and 'flash' in precedence }}",
                                        "flash_override": "{{ True if precedence is defined and 'flash-override' in precedence }}",
                                        "immediate": "{{ True if precedence is defined and 'immediate' in precedence }}",
                                        "internet": "{{ True if precedence is defined and 'internet' in precedence }}",
                                        "network": "{{ True if precedence is defined and 'network' in precedence }}",
                                        "priority": "{{ True if precedence is defined and 'priority' in precedence }}",
                                        "routine": "{{ True if precedence is defined and 'routine' in precedence }}",
                                    },
                                    "qos_group": "{{ qos_group.split('qos-group ')[1] if qos_group is defined }}",
                                    "tos": {
                                        "set": "{{ True if tos is defined and tos.split(' ')|length == 1 }}",
                                        "max_reliability": "{{ True if tos is defined and 'max-reliability' in tos }}",
                                        "max_throughput": "{{ True if tos is defined and 'max-throughput' in tos }}",
                                        "min_delay": "{{ True if tos is defined and 'min-delay' in tos }}",
                                        "min_monetary_cost": "{{ True if tos is defined and 'min-monetary-cost' in tos }}",
                                        "normal": "{{ True if tos is defined and 'normal' in tos }}",
                                    },
                                    "vrf": {
                                        "name": "{{ vrf.split(' ')[1] if vrf is defined }}",
                                        "address": "{{ vrf.split('next-hop ')[1] if vrf is defined and 'verify-availability' not in vrf }}",
                                        "verify_availability": {
                                            "set": "{{ True if vrf is defined and 'track' not in vrf }}",
                                            "address": "{{ vrf.split(' ')[4] if vrf is defined and 'track' in vrf }}",
                                            "sequence": "{{ vrf.split(' ')[5] if vrf is defined and 'track' in vrf }}",
                                            "track": "{{ vrf.split('track ')[1] if vrf is defined and 'track' in vrf }}",
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "set.ipv6",
            "getval": re.compile(
                r"""
                \s+set*
                \s*(?P<ipv6>ipv6)*
                \s*(?P<address>address\sprefix-list\s\S+)*
                \s*(?P<default>default\snext-hop\s(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))\S+)*
                \s*(?P<global>global\snext-hop\sverify-availability\s(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))\S+\s\d+\strack\s\d+)*
                \s*(?P<precedence>precedence\s\d+)*
                \s*(?P<vrf>vrf\s\S+\snext-hop\sverify-availability\s(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))\S+\s\d+\strack\s\d+)*
                \s*(?P<next_hop>next-hop\s\S.*)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_route_map_set_ipv6,
            "compval": "set",
            "result": {
                "{{ route_map }}": {
                    "{{ action|d() + '_' + sequence|d() }}": {
                        "entries": {
                            "set": {
                                "ipv6": {
                                    "address": "{{ address.split('address prefix-list ')[1] if address is defined }}",
                                    "default": "{{ default.split('default next-hop ')[1] if default is defined }}",
                                    "global_route": {
                                        "verify_availability": {
                                            "address": "{{ global.split(' ')[3] if global is defined and 'verify-availability' in global }}",
                                            "sequence": "{{ global.split(' ')[4] if global is defined and 'verify-availability' in global }}",
                                            "track": "{{ global.split('track ')[1] if global is defined and 'verify-availability' in global }}",
                                        },
                                        "address": "{{ global.split(' ')[2] if global is defined and 'verify-availability' not in global }}",
                                    },
                                    "next_hop": {
                                        "address": "{{ next_hop.split('next-hop ')[1] if next_hop is defined and\
                                                 next_hop.split(' ')|length == 2 and 'peer-address' not in next_hop }}",
                                        "encapsulate": "{{ next_hop.split('next-hop encapsulate l3vpn ')[1] if next_hop is defined and\
                                                 'encapsulate' in next_hop}}",
                                        "peer_address": "{{ True if next_hop is defined and next_hop.split(' ')|length == 2 and 'peer-address' in next_hop }}",
                                        "recursive": "{{ next_hop.split('next-hop recursive ')[1] if next_hop is defined }}",
                                    },
                                    "precedence": "{{ precedence.split(' ')[1] if precedence is defined }}",
                                    "vrf": {
                                        "name": "{{ vrf.split(' ')[1] if vrf is defined }}",
                                        "verify_availability": {
                                            "address": "{{ vrf.split(' ')[4] if vrf is defined and 'verify-availability' in vrf }}",
                                            "sequence": "{{ vrf.split(' ')[5] if vrf is defined and 'verify-availability' in vrf }}",
                                            "track": "{{ vrf.split('track ')[1] if vrf is defined and 'verify-availability' in vrf }}",
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    ]



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright: (c) 2018, Ansible Project
# Copyright: (c) 2018, Anthony Bond <ajbond2005@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type


DOCUMENTATION = """
---
module: digital_ocean_firewall
short_description: Manage cloud firewalls within DigitalOcean
description:
    - This module can be used to add or remove firewalls on the DigitalOcean cloud platform.
author:
    - Anthony Bond (@BondAnthony)
    - Lucas Basquerotto (@lucasbasquerotto)
version_added: "1.1.0"
options:
  name:
    type: str
    description:
     - Name of the firewall rule to create or manage
    required: true
  state:
    type: str
    choices: ['present', 'absent']
    default: present
    description:
      - Assert the state of the firewall rule. Set to 'present' to create or update and 'absent' to remove.
  droplet_ids:
    type: list
    elements: str
    description:
     - List of droplet ids to be assigned to the firewall
    required: false
  tags:
    type: list
    elements: str
    description:
      - List of tags to be assigned to the firewall
    required: false
  inbound_rules:
    type: list
    elements: dict
    description:
      - Firewall rules specifically targeting inbound network traffic into DigitalOcean
    required: false
    suboptions:
      protocol:
        type: str
        choices: ['udp', 'tcp', 'icmp']
        default: tcp
        description:
          - Network protocol to be accepted.
        required: false
      ports:
        type: str
        description:
          - The ports on which traffic will be allowed, single, range, or all
        required: true
      sources:
        type: dict
        description:
          - Dictionary of locations from which inbound traffic will be accepted
        required: true
        suboptions:
          addresses:
            type: list
            elements: str
            description:
              - List of strings containing the IPv4 addresses, IPv6 addresses, IPv4 CIDRs,
                and/or IPv6 CIDRs to which the firewall will allow traffic
            required: false
          droplet_ids:
            type: list
            elements: str
            description:
              - List of integers containing the IDs of the Droplets to which the firewall will allow traffic
            required: false
          load_balancer_uids:
            type: list
            elements: str
            description:
              - List of strings containing the IDs of the Load Balancers to which the firewall will allow traffic
            required: false
          tags:
            type: list
            elements: str
            description:
              - List of strings containing the names of Tags corresponding to groups of Droplets to
                which the Firewall will allow traffic
            required: false
  outbound_rules:
    type: list
    elements: dict
    description:
      - Firewall rules specifically targeting outbound network traffic from DigitalOcean
    required: false
    suboptions:
      protocol:
        type: str
        choices: ['udp', 'tcp', 'icmp']
        default: tcp
        description:
          - Network protocol to be accepted.
        required: false
      ports:
        type: str
        description:
          - The ports on which traffic will be allowed, single, range, or all
        required: true
      destinations:
        type: dict
        description:
          - Dictionary of locations from which outbound traffic will be allowed
        required: true
        suboptions:
          addresses:
            type: list
            elements: str
            description:
              - List of strings containing the IPv4 addresses, IPv6 addresses, IPv4 CIDRs,
                and/or IPv6 CIDRs to which the firewall will allow traffic
            required: false
          droplet_ids:
            type: list
            elements: str
            description:
              - List of integers containing the IDs of the Droplets to which the firewall will allow traffic
            required: false
          load_balancer_uids:
            type: list
            elements: str
            description:
              - List of strings containing the IDs of the Load Balancers to which the firewall will allow traffic
            required: false
          tags:
            type: list
            elements: str
            description:
              - List of strings containing the names of Tags corresponding to groups of Droplets to
                which the Firewall will allow traffic
            required: false
extends_documentation_fragment:
  - community.digitalocean.digital_ocean.documentation
"""

EXAMPLES = """
# Allows tcp connections to port 22 (SSH) from specific sources
# Allows tcp connections to ports 80 and 443 from any source
# Allows outbound access to any destination for protocols tcp, udp and icmp
# The firewall rules will be applied to any droplets with the tag "sample"
- name: Create a Firewall named my-firewall
  digital_ocean_firewall:
    name: my-firewall
    state: present
    inbound_rules:
      - protocol: "tcp"
        ports: "22"
        sources:
          addresses: ["1.2.3.4"]
          droplet_ids: ["my_droplet_id_1", "my_droplet_id_2"]
          load_balancer_uids: ["my_lb_id_1", "my_lb_id_2"]
          tags: ["tag_1", "tag_2"]
      - protocol: "tcp"
        ports: "80"
        sources:
          addresses: ["0.0.0.0/0", "::/0"]
      - protocol: "tcp"
        ports: "443"
        sources:
          addresses: ["0.0.0.0/0", "::/0"]
    outbound_rules:
      - protocol: "tcp"
        ports: "1-65535"
        destinations:
          addresses: ["0.0.0.0/0", "::/0"]
      - protocol: "udp"
        ports: "1-65535"
        destinations:
          addresses: ["0.0.0.0/0", "::/0"]
      - protocol: "icmp"
        ports: "1-65535"
        destinations:
          addresses: ["0.0.0.0/0", "::/0"]
    droplet_ids: []
    tags: ["sample"]
"""

RETURN = """
data:
    description: DigitalOcean firewall resource
    returned: success
    type: dict
    sample: {
        "created_at": "2020-08-11T18:41:30Z",
        "droplet_ids": [],
        "id": "7acd6ee2-257b-434f-8909-709a5816d4f9",
        "inbound_rules": [
            {
                "ports": "443",
                "protocol": "tcp",
                "sources": {
                  "addresses": [
                      "1.2.3.4"
                  ],
                  "droplet_ids": [
                      "my_droplet_id_1",
                      "my_droplet_id_2"
                  ],
                  "load_balancer_uids": [
                      "my_lb_id_1",
                      "my_lb_id_2"
                  ],
                  "tags": [
                      "tag_1",
                      "tag_2"
                  ]
                }
            },
            {
                "sources": {
                    "addresses": [
                        "0.0.0.0/0",
                        "::/0"
                    ]
                },
                "ports": "80",
                "protocol": "tcp"
            },
            {
                "sources": {
                    "addresses": [
                        "0.0.0.0/0",
                        "::/0"
                    ]
                },
                "ports": "443",
                "protocol": "tcp"
            }
        ],
        "name": "my-firewall",
        "outbound_rules": [
            {
                "destinations": {
                    "addresses": [
                        "0.0.0.0/0",
                        "::/0"
                    ]
                },
                "ports": "1-65535",
                "protocol": "tcp"
            },
            {
                "destinations": {
                    "addresses": [
                        "0.0.0.0/0",
                        "::/0"
                    ]
                },
                "ports": "1-65535",
                "protocol": "udp"
            },
            {
                "destinations": {
                    "addresses": [
                        "0.0.0.0/0",
                        "::/0"
                    ]
                },
                "ports": "1-65535",
                "protocol": "icmp"
            }
        ],
        "pending_changes": [],
        "status": "succeeded",
        "tags": ["sample"]
    }
"""

from traceback import format_exc
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.digitalocean.plugins.module_utils.digital_ocean import (
    DigitalOceanHelper,
)
from ansible.module_utils._text import to_native

address_spec = dict(
    addresses=dict(type="list", elements="str", required=False),
    droplet_ids=dict(type="list", elements="str", required=False),
    load_balancer_uids=dict(type="list", elements="str", required=False),
    tags=dict(type="list", elements="str", required=False),
)

inbound_spec = dict(
    protocol=dict(type="str", choices=["udp", "tcp", "icmp"], default="tcp"),
    ports=dict(type="str", required=True),
    sources=dict(type="dict", required=True, options=address_spec),
)

outbound_spec = dict(
    protocol=dict(type="str", choices=["udp", "tcp", "icmp"], default="tcp"),
    ports=dict(type="str", required=True),
    destinations=dict(type="dict", required=True, options=address_spec),
)


class DOFirewall(object):
    def __init__(self, module):
        self.rest = DigitalOceanHelper(module)
        self.module = module
        self.name = self.module.params.get("name")
        self.baseurl = "firewalls"
        self.firewalls = self.get_firewalls()

    def get_firewalls(self):
        base_url = self.baseurl + "?"
        response = self.rest.get("%s" % base_url)
        status_code = response.status_code
        status_code_success = 200

        if status_code != status_code_success:
            error = response.json
            info = response.info

            if error:
                error.update({"status_code": status_code})
                error.update({"status_code_success": status_code_success})
                self.module.fail_json(msg=error)
            elif info:
                info.update({"status_code_success": status_code_success})
                self.module.fail_json(msg=info)
            else:
                msg_error = "Failed to retrieve firewalls from DigitalOcean"
                self.module.fail_json(
                    msg=msg_error
                    + " (url="
                    + self.rest.baseurl
                    + "/"
                    + self.baseurl
                    + ", status="
                    + str(status_code or "")
                    + " - expected:"
                    + str(status_code_success)
                    + ")"
                )

        return self.rest.get_paginated_data(
            base_url=base_url, data_key_name="firewalls"
        )

    def get_firewall_by_name(self):
        rule = {}
        for firewall in self.firewalls:
            if firewall["name"] == self.name:
                rule.update(firewall)
                return rule
        return None

    def ordered(self, obj):
        if isinstance(obj, dict):
            return sorted((k, self.ordered(v)) for k, v in obj.items())
        if isinstance(obj, list):
            return sorted(self.ordered(x) for x in obj)
        else:
            return obj

    def fill_protocol_defaults(self, obj):
        if obj.get("protocol") is None:
            obj["protocol"] = "tcp"

        return obj

    def fill_source_and_destination_defaults_inner(self, obj):
        addresses = obj.get("addresses") or []

        droplet_ids = obj.get("droplet_ids") or []
        droplet_ids = [str(droplet_id) for droplet_id in droplet_ids]

        load_balancer_uids = obj.get("load_balancer_uids") or []
        load_balancer_uids = [str(uid) for uid in load_balancer_uids]

        tags = obj.get("tags") or []

        data = {
            "addresses": addresses,
            "droplet_ids": droplet_ids,
            "load_balancer_uids": load_balancer_uids,
            "tags": tags,
        }

        return data

    def fill_sources_and_destinations_defaults(self, obj, prop):
        value = obj.get(prop)

        if value is None:
            value = {}
        else:
            value = self.fill_source_and_destination_defaults_inner(value)

        obj[prop] = value

        return obj

    def fill_data_defaults(self, obj):
        inbound_rules = obj.get("inbound_rules")

        if inbound_rules is None:
            inbound_rules = []
        else:
            inbound_rules = [self.fill_protocol_defaults(x) for x in inbound_rules]
            inbound_rules = [
                self.fill_sources_and_destinations_defaults(x, "sources")
                for x in inbound_rules
            ]

        outbound_rules = obj.get("outbound_rules")

        if outbound_rules is None:
            outbound_rules = []
        else:
            outbound_rules = [self.fill_protocol_defaults(x) for x in outbound_rules]
            outbound_rules = [
                self.fill_sources_and_destinations_defaults(x, "destinations")
                for x in outbound_rules
            ]

        droplet_ids = obj.get("droplet_ids") or []
        droplet_ids = [str(droplet_id) for droplet_id in droplet_ids]

        tags = obj.get("tags") or []

        data = {
            "name": obj.get("name"),
            "inbound_rules": inbound_rules,
            "outbound_rules": outbound_rules,
            "droplet_ids": droplet_ids,
            "tags": tags,
        }

        return data

    def data_to_compare(self, obj):
        return self.ordered(self.fill_data_defaults(obj))

    def update(self, obj, id):
        if id is None:
            status_code_success = 202
            resp = self.rest.post(path=self.baseurl, data=obj)
        else:
            status_code_success = 200
            resp = self.rest.put(path=self.baseurl + "/" + id, data=obj)
        status_code = resp.status_code
        if status_code != status_code_success:
            error = resp.json
            error.update(
                {
                    "context": "error when trying to "
                    + ("create" if (id is None) else "update")
                    + " firewalls"
                }
            )
            error.update({"status_code": status_code})
            error.update({"status_code_success": status_code_success})
            self.module.fail_json(msg=error)
        self.module.exit_json(changed=True, data=resp.json["firewall"])

    def create(self):
        rule = self.get_firewall_by_name()
        data = {
            "name": self.module.params.get("name"),
            "inbound_rules": self.module.params.get("inbound_rules"),
            "outbound_rules": self.module.params.get("outbound_rules"),
            "droplet_ids": self.module.params.get("droplet_ids"),
            "tags": self.module.params.get("tags"),
        }
        if rule is None:
            self.update(data, None)
        else:
            rule_data = {
                "name": rule.get("name"),
                "inbound_rules": rule.get("inbound_rules"),
                "outbound_rules": rule.get("outbound_rules"),
                "droplet_ids": rule.get("droplet_ids"),
                "tags": rule.get("tags"),
            }

            user_data = {
                "name": data.get("name"),
                "inbound_rules": data.get("inbound_rules"),
                "outbound_rules": data.get("outbound_rules"),
                "droplet_ids": data.get("droplet_ids"),
                "tags": data.get("tags"),
            }

            if self.data_to_compare(user_data) == self.data_to_compare(rule_data):
                self.module.exit_json(changed=False, data=rule)
            else:
                self.update(data, rule.get("id"))

    def destroy(self):
        rule = self.get_firewall_by_name()
        if rule is None:
            self.module.exit_json(changed=False, data="Firewall does not exist")
        else:
            endpoint = self.baseurl + "/" + rule["id"]
            resp = self.rest.delete(path=endpoint)
            status_code = resp.status_code
            if status_code != 204:
                self.module.fail_json(msg="Failed to delete firewall")
            self.module.exit_json(
                changed=True,
                data="Deleted firewall rule: {0} - {1}".format(
                    rule["name"], rule["id"]
                ),
            )


def core(module):
    state = module.params.get("state")
    firewall = DOFirewall(module)

    if state == "present":
        firewall.create()
    elif state == "absent":
        firewall.destroy()


def main():
    argument_spec = DigitalOceanHelper.digital_ocean_argument_spec()
    argument_spec.update(
        name=dict(type="str", required=True),
        state=dict(type="str", choices=["present", "absent"], default="present"),
        droplet_ids=dict(type="list", elements="str", required=False),
        tags=dict(type="list", elements="str", required=False),
        inbound_rules=dict(
            type="list", elements="dict", options=inbound_spec, required=False
        ),
        outbound_rules=dict(
            type="list", elements="dict", options=outbound_spec, required=False
        ),
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        required_if=[("state", "present", ["inbound_rules", "outbound_rules"])],
    )

    try:
        core(module)
    except Exception as e:
        module.fail_json(msg=to_native(e), exception=format_exc())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# (c) 2022 Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from ansible.module_utils._text import to_native


class AnsibleAWSError(Exception):
    def __str__(self):
        if self.exception and self.message:
            return f"{self.message}: {to_native(self.exception)}"

        return super().__str__()

    def __init__(self, message=None, exception=None, **kwargs):
        if not message and not exception:
            super().__init__()
        elif not message:
            super().__init__(exception)
        else:
            super().__init__(message)

        self.exception = exception
        self.message = message

        # In places where passing more information to module.fail_json would be helpful
        # store the extra info.  Other plugin types have to raise the correct exception
        # such as AnsibleLookupError, so can't easily consume this.
        self.kwargs = kwargs or {}


class AnsibleBotocoreError(AnsibleAWSError):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: networks_wireless_ssids_firewall_l3_firewall_rules
short_description: Resource module for networks _wireless _ssids _firewall l3firewallrules
description:
- Manage operation update of the resource networks _wireless _ssids _firewall l3firewallrules.
- Update the L3 firewall rules of an SSID on an MR network.
version_added: '2.16.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  allowLanAccess:
    description: Allow wireless client access to local LAN (boolean value - true allows
      access and false denies access) (optional).
    type: bool
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  number:
    description: Number path parameter.
    type: str
  rules:
    description: An ordered array of the firewall rules for this SSID (not including
      the local LAN access rule or the default rule).
    elements: dict
    suboptions:
      comment:
        description: Description of the rule (optional).
        type: str
      destCidr:
        description: Comma-separated list of destination IP address(es) (in IP or CIDR
          notation), fully-qualified domain names (FQDN) or 'any'.
        type: str
      destPort:
        description: Comma-separated list of destination port(s) (integer in the range
          1-65535), or 'any'.
        type: str
      policy:
        description: '''allow'' or ''deny'' traffic specified by this rule.'
        type: str
      protocol:
        description: The type of protocol (must be 'tcp', 'udp', 'icmp', 'icmp6' or
          'any').
        type: str
    type: list
requirements:
- meraki >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco Meraki documentation for wireless updateNetworkWirelessSsidFirewallL3FirewallRules
  description: Complete reference of the updateNetworkWirelessSsidFirewallL3FirewallRules API.
  link: https://developer.cisco.com/meraki/api-v1/#!update-network-wireless-ssid-firewall-l3-firewall-rules
notes:
  - SDK Method used are
    wireless.Wireless.update_network_wireless_ssid_firewall_l3_firewall_rules,

  - Paths used are
    put /networks/{networkId}/wireless/ssids/{number}/firewall/l3FirewallRules,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.networks_wireless_ssids_firewall_l3_firewall_rules:
    meraki_api_key: "{{meraki_api_key}}"
    meraki_base_url: "{{meraki_base_url}}"
    meraki_single_request_timeout: "{{meraki_single_request_timeout}}"
    meraki_certificate_path: "{{meraki_certificate_path}}"
    meraki_requests_proxy: "{{meraki_requests_proxy}}"
    meraki_wait_on_rate_limit: "{{meraki_wait_on_rate_limit}}"
    meraki_nginx_429_retry_wait_time: "{{meraki_nginx_429_retry_wait_time}}"
    meraki_action_batch_retry_wait_time: "{{meraki_action_batch_retry_wait_time}}"
    meraki_retry_4xx_error: "{{meraki_retry_4xx_error}}"
    meraki_retry_4xx_error_wait_time: "{{meraki_retry_4xx_error_wait_time}}"
    meraki_maximum_retries: "{{meraki_maximum_retries}}"
    meraki_output_log: "{{meraki_output_log}}"
    meraki_log_file_prefix: "{{meraki_log_file_prefix}}"
    meraki_log_path: "{{meraki_log_path}}"
    meraki_print_console: "{{meraki_print_console}}"
    meraki_suppress_logging: "{{meraki_suppress_logging}}"
    meraki_simulate: "{{meraki_simulate}}"
    meraki_be_geo_id: "{{meraki_be_geo_id}}"
    meraki_use_iterator_for_get_pages: "{{meraki_use_iterator_for_get_pages}}"
    meraki_inherit_logging_config: "{{meraki_inherit_logging_config}}"
    state: present
    networkId: string
    number: string
    rules:
    - comment: Allow TCP traffic to subnet with HTTP servers.
      destCidr: 192.168.1.0/24
      destPort: '443'
      policy: allow
      protocol: tcp

"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {}
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: elb_application_lb_info
version_added: 5.0.0
short_description: Gather information about Application Load Balancers in AWS
description:
  - Gather information about Application Load Balancers in AWS
  - This module was originally added to C(community.aws) in release 1.0.0.
author:
  - Rob White (@wimnat)
options:
  load_balancer_arns:
    description:
      - The Amazon Resource Names (ARN) of the load balancers. You can specify up to 20 load balancers in a single call.
    required: false
    type: list
    elements: str
  names:
    description:
      - The names of the load balancers.
    required: false
    type: list
    elements: str
  include_attributes:
    description:
      - Whether or not to include load balancer attributes in the response.
    required: false
    type: bool
    default: true
    version_added: 7.0.0
  include_listeners:
    description:
      - Whether or not to include load balancer listeners in the response.
    required: false
    type: bool
    default: true
    version_added: 7.0.0
  include_listener_rules:
    description:
      - Whether or not to include load balancer listener rules in the response.
      - Implies I(include_listeners=true)
    required: false
    type: bool
    default: true
    version_added: 7.0.0

extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Gather information about all ALBs
  amazon.aws.elb_application_lb_info:

# Equivalent to aws elbv2 describe-load-balancers
- name: Gather minimal information about all ALBs
  amazon.aws.elb_application_lb_info:
    include_attributes: false
    include_listeners: false
    include_listener_rules: false

- name: Gather information about a particular ALB given its ARN
  amazon.aws.elb_application_lb_info:
    load_balancer_arns:
      - "arn:aws:elasticloadbalancing:ap-southeast-2:123456789012:loadbalancer/app/my-alb/aabbccddeeff"

- name: Gather information about ALBs named 'alb1' and 'alb2'
  amazon.aws.elb_application_lb_info:
    names:
      - alb1
      - alb2

- name: Gather information about specific ALB
  amazon.aws.elb_application_lb_info:
    names: "alb-name"
    region: "aws-region"
  register: alb_info
- ansible.builtin.debug:
    var: alb_info
"""

RETURN = r"""
load_balancers:
    description: a list of load balancers
    returned: always
    type: complex
    contains:
        access_logs_s3_bucket:
            description: The name of the S3 bucket for the access logs.
            returned: when include_attributes is true
            type: str
            sample: "mys3bucket"
        access_logs_s3_enabled:
            description: Indicates whether access logs stored in Amazon S3 are enabled.
            returned: when include_attributes is true
            type: bool
            sample: true
        access_logs_s3_prefix:
            description: The prefix for the location in the S3 bucket.
            returned: when include_attributes is true
            type: str
            sample: "my/logs"
        availability_zones:
            description: The Availability Zones for the load balancer.
            type: list
            sample: [{ "load_balancer_addresses": [], "subnet_id": "subnet-aabbccddff", "zone_name": "ap-southeast-2a" }]
        canonical_hosted_zone_id:
            description: The ID of the Amazon Route 53 hosted zone associated with the load balancer.
            type: str
            sample: "ABCDEF12345678"
        created_time:
            description: The date and time the load balancer was created.
            type: str
            sample: "2015-02-12T02:14:02+00:00"
        deletion_protection_enabled:
            description: Indicates whether deletion protection is enabled.
            returned: when include_attributes is true
            type: bool
            sample: true
        dns_name:
            description: The public DNS name of the load balancer.
            type: str
            sample: "internal-my-alb-123456789.ap-southeast-2.elb.amazonaws.com"
        idle_timeout_timeout_seconds:
            description: The idle timeout value, in seconds.
            returned: when include_attributes is true
            type: int
            sample: 60
        ip_address_type:
            description: The type of IP addresses used by the subnets for the load balancer.
            type: str
            sample: "ipv4"
        listeners:
            description: Information about the listeners.
            returned: when include_listeners or include_listener_rules is true
            type: complex
            contains:
                listener_arn:
                    description: The Amazon Resource Name (ARN) of the listener.
                    type: str
                    sample: ""
                load_balancer_arn:
                    description: The Amazon Resource Name (ARN) of the load balancer.
                    type: str
                    sample: ""
                port:
                    description: The port on which the load balancer is listening.
                    type: int
                    sample: 80
                protocol:
                    description: The protocol for connections from clients to the load balancer.
                    type: str
                    sample: "HTTPS"
                rules:
                    description: List of listener rules.
                    returned: when include_listener_rules is true
                    type: list
                    sample: ""
                certificates:
                    description: The SSL server certificate.
                    type: complex
                    contains:
                        certificate_arn:
                            description: The Amazon Resource Name (ARN) of the certificate.
                            type: str
                            sample: ""
                ssl_policy:
                    description: The security policy that defines which ciphers and protocols are supported.
                    type: str
                    sample: ""
                default_actions:
                    description: The default actions for the listener.
                    type: str
                    contains:
                        type:
                            description: The type of action.
                            type: str
                            sample: ""
                        target_group_arn:
                            description: The Amazon Resource Name (ARN) of the target group.
                            type: str
                            sample: ""
        load_balancer_arn:
            description: The Amazon Resource Name (ARN) of the load balancer.
            type: str
            sample: "arn:aws:elasticloadbalancing:ap-southeast-2:123456789012:loadbalancer/app/my-alb/001122334455"
        load_balancer_name:
            description: The name of the load balancer.
            type: str
            sample: "my-alb"
        load_balancing_cross_zone_enabled:
            description: Indicates whether or not cross-zone load balancing is enabled.
            returned: when include_attributes is true
            type: bool
            sample: true
        routing_http2_enabled:
            description: Indicates whether HTTP/2 is enabled.
            returned: when include_attributes is true
            type: bool
            sample: true
        routing_http_desync_mitigation_mode:
            description: Determines how the load balancer handles requests that might pose a security risk to an application.
            returned: when include_attributes is true
            type: str
            sample: "defensive"
        routing_http_drop_invalid_header_fields_enabled:
            description: Indicates whether HTTP headers with invalid header fields are removed by the load balancer (true) or routed to targets (false).
            returned: when include_attributes is true
            type: bool
            sample: false
        routing_http_x_amzn_tls_version_and_cipher_suite_enabled:
            description: Indicates whether the two headers are added to the client request before sending it to the target.
            returned: when include_attributes is true
            type: bool
            sample: false
        routing_http_xff_client_port_enabled:
            description: Indicates whether the X-Forwarded-For header should preserve the source port that the client used to connect to the load balancer.
            returned: when include_attributes is true
            type: bool
            sample: false
        scheme:
            description: Internet-facing or internal load balancer.
            type: str
            sample: "internal"
        security_groups:
            description: The IDs of the security groups for the load balancer.
            type: list
            sample: ['sg-0011223344']
        state:
            description: The state of the load balancer.
            type: dict
            sample: {'code': 'active'}
        tags:
            description: The tags attached to the load balancer.
            type: dict
            sample: {
                'Tag': 'Example'
            }
        type:
            description: The type of load balancer.
            type: str
            sample: "application"
        vpc_id:
            description: The ID of the VPC for the load balancer.
            type: str
            sample: "vpc-0011223344"
        waf_fail_open_enabled:
            description: Indicates whether to allow a AWS WAF-enabled load balancer to route requests to targets
                if it is unable to forward the request to AWS WAF.
            returned: when include_attributes is true
            type: bool
            sample: false
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict


@AWSRetry.jittered_backoff(retries=10)
def get_paginator(connection, **kwargs):
    paginator = connection.get_paginator("describe_load_balancers")
    return paginator.paginate(**kwargs).build_full_result()


def get_alb_listeners(connection, module, alb_arn):
    try:
        return connection.describe_listeners(
            aws_retry=True,
            LoadBalancerArn=alb_arn,
        )["Listeners"]
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to describe alb listeners")


def get_listener_rules(connection, module, listener_arn):
    try:
        return connection.describe_rules(
            aws_retry=True,
            ListenerArn=listener_arn,
        )["Rules"]
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to describe listener rules")


def get_load_balancer_attributes(connection, module, load_balancer_arn):
    try:
        attributes = connection.describe_load_balancer_attributes(
            aws_retry=True,
            LoadBalancerArn=load_balancer_arn,
        )["Attributes"]
        load_balancer_attributes = boto3_tag_list_to_ansible_dict(attributes)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to describe load balancer attributes")

    # Replace '.' with '_' in attribute key names to make it more Ansibley
    for k, v in list(load_balancer_attributes.items()):
        load_balancer_attributes[k.replace(".", "_")] = v
        del load_balancer_attributes[k]

    return load_balancer_attributes


def get_load_balancer_tags(connection, module, load_balancer_arn):
    try:
        tag_descriptions = connection.describe_tags(
            aws_retry=True,
            ResourceArns=[load_balancer_arn],
        )["TagDescriptions"]
        return boto3_tag_list_to_ansible_dict(tag_descriptions[0]["Tags"])
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to describe load balancer tags")


def list_load_balancers(connection, module):
    load_balancer_arns = module.params.get("load_balancer_arns")
    names = module.params.get("names")
    include_attributes = module.params.get("include_attributes")
    include_listeners = module.params.get("include_listeners")
    include_listener_rules = module.params.get("include_listener_rules")

    try:
        if not load_balancer_arns and not names:
            load_balancers = get_paginator(connection)
        if load_balancer_arns:
            load_balancers = get_paginator(connection, LoadBalancerArns=load_balancer_arns)
        if names:
            load_balancers = get_paginator(connection, Names=names)
    except is_boto3_error_code("LoadBalancerNotFound"):
        module.exit_json(load_balancers=[])
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Failed to list load balancers")

    for load_balancer in load_balancers["LoadBalancers"]:
        # Get the attributes for each alb
        if include_attributes:
            load_balancer.update(get_load_balancer_attributes(connection, module, load_balancer["LoadBalancerArn"]))

        # Get the listeners for each alb
        if include_listeners or include_listener_rules:
            load_balancer["listeners"] = get_alb_listeners(connection, module, load_balancer["LoadBalancerArn"])

        # For each listener, get listener rules
        if include_listener_rules:
            for listener in load_balancer["listeners"]:
                listener["rules"] = get_listener_rules(connection, module, listener["ListenerArn"])

    # Turn the boto3 result in to ansible_friendly_snaked_names
    snaked_load_balancers = [
        camel_dict_to_snake_dict(load_balancer) for load_balancer in load_balancers["LoadBalancers"]
    ]

    # Get tags for each load balancer
    for snaked_load_balancer in snaked_load_balancers:
        snaked_load_balancer["tags"] = get_load_balancer_tags(
            connection, module, snaked_load_balancer["load_balancer_arn"]
        )

    module.exit_json(load_balancers=snaked_load_balancers)


def main():
    argument_spec = dict(
        load_balancer_arns=dict(type="list", elements="str"),
        names=dict(type="list", elements="str"),
        include_attributes=dict(default=True, type="bool"),
        include_listeners=dict(default=True, type="bool"),
        include_listener_rules=dict(default=True, type="bool"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        mutually_exclusive=[["load_balancer_arns", "names"]],
        supports_check_mode=True,
    )

    try:
        connection = module.client("elbv2", retry_decorator=AWSRetry.jittered_backoff(retries=10))
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    list_load_balancers(connection, module)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Allen Sanabria, <asanabria@linuxdynasty.org>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


"""
This module adds shared support for generic cloud modules

In order to use this module, include it as part of a custom
module as shown below.

from ansible.module_utils.cloud import CloudRetry

The 'cloud' module provides the following common classes:

    * CloudRetry
        - The base class to be used by other cloud providers, in order to
          provide a backoff/retry decorator based on status codes.

        - Example using the AWSRetry class which inherits from CloudRetry.

          @AWSRetry.exponential_backoff(retries=10, delay=3)
          get_ec2_security_group_ids_from_names()

          @AWSRetry.jittered_backoff()
          get_ec2_security_group_ids_from_names()

"""
import random
from functools import wraps
import syslog
import time


def _exponential_backoff(retries=10, delay=2, backoff=2, max_delay=60):
    """ Customizable exponential backoff strategy.
    Args:
        retries (int): Maximum number of times to retry a request.
        delay (float): Initial (base) delay.
        backoff (float): base of the exponent to use for exponential
            backoff.
        max_delay (int): Optional. If provided each delay generated is capped
            at this amount. Defaults to 60 seconds.
    Returns:
        Callable that returns a generator. This generator yields durations in
        seconds to be used as delays for an exponential backoff strategy.
    Usage:
        >>> backoff = _exponential_backoff()
        >>> backoff
        <function backoff_backoff at 0x7f0d939facf8>
        >>> list(backoff())
        [2, 4, 8, 16, 32, 60, 60, 60, 60, 60]
    """
    def backoff_gen():
        for retry in range(0, retries):
            sleep = delay * backoff ** retry
            yield sleep if max_delay is None else min(sleep, max_delay)
    return backoff_gen


def _full_jitter_backoff(retries=10, delay=3, max_delay=60, _random=random):
    """ Implements the "Full Jitter" backoff strategy described here
    https://www.awsarchitectureblog.com/2015/03/backoff.html
    Args:
        retries (int): Maximum number of times to retry a request.
        delay (float): Approximate number of seconds to sleep for the first
            retry.
        max_delay (int): The maximum number of seconds to sleep for any retry.
            _random (random.Random or None): Makes this generator testable by
            allowing developers to explicitly pass in the a seeded Random.
    Returns:
        Callable that returns a generator. This generator yields durations in
        seconds to be used as delays for a full jitter backoff strategy.
    Usage:
        >>> backoff = _full_jitter_backoff(retries=5)
        >>> backoff
        <function backoff_backoff at 0x7f0d939facf8>
        >>> list(backoff())
        [3, 6, 5, 23, 38]
        >>> list(backoff())
        [2, 1, 6, 6, 31]
    """
    def backoff_gen():
        for retry in range(0, retries):
            yield _random.randint(0, min(max_delay, delay * 2 ** retry))
    return backoff_gen


class CloudRetry(object):
    """ CloudRetry can be used by any cloud provider, in order to implement a
        backoff algorithm/retry effect based on Status Code from Exceptions.
    """
    # This is the base class of the exception.
    # AWS Example botocore.exceptions.ClientError
    base_class = None

    @staticmethod
    def status_code_from_exception(error):
        """ Return the status code from the exception object
        Args:
            error (object): The exception itself.
        """
        pass

    @staticmethod
    def found(response_code, catch_extra_error_codes=None):
        """ Return True if the Response Code to retry on was found.
        Args:
            response_code (str): This is the Response Code that is being matched against.
        """
        pass

    @classmethod
    def _backoff(cls, backoff_strategy, catch_extra_error_codes=None):
        """ Retry calling the Cloud decorated function using the provided
        backoff strategy.
        Args:
            backoff_strategy (callable): Callable that returns a generator. The
            generator should yield sleep times for each retry of the decorated
            function.
        """
        def deco(f):
            @wraps(f)
            def retry_func(*args, **kwargs):
                for delay in backoff_strategy():
                    try:
                        return f(*args, **kwargs)
                    except Exception as e:
                        if isinstance(e, cls.base_class):  # pylint: disable=isinstance-second-argument-not-valid-type
                            response_code = cls.status_code_from_exception(e)
                            if cls.found(response_code, catch_extra_error_codes):
                                msg = "{0}: Retrying in {1} seconds...".format(str(e), delay)
                                syslog.syslog(syslog.LOG_INFO, msg)
                                time.sleep(delay)
                            else:
                                # Return original exception if exception is not a ClientError
                                raise e
                        else:
                            # Return original exception if exception is not a ClientError
                            raise e
                return f(*args, **kwargs)

            return retry_func  # true decorator

        return deco

    @classmethod
    def exponential_backoff(cls, retries=10, delay=3, backoff=2, max_delay=60, catch_extra_error_codes=None):
        """
        Retry calling the Cloud decorated function using an exponential backoff.

        Kwargs:
            retries (int): Number of times to retry a failed request before giving up
                default=10
            delay (int or float): Initial delay between retries in seconds
                default=3
            backoff (int or float): backoff multiplier e.g. value of 2 will
                double the delay each retry
                default=1.1
            max_delay (int or None): maximum amount of time to wait between retries.
                default=60
        """
        return cls._backoff(_exponential_backoff(
            retries=retries, delay=delay, backoff=backoff, max_delay=max_delay), catch_extra_error_codes)

    @classmethod
    def jittered_backoff(cls, retries=10, delay=3, max_delay=60, catch_extra_error_codes=None):
        """
        Retry calling the Cloud decorated function using a jittered backoff
        strategy. More on this strategy here:

        https://www.awsarchitectureblog.com/2015/03/backoff.html

        Kwargs:
            retries (int): Number of times to retry a failed request before giving up
                default=10
            delay (int): Initial delay between retries in seconds
                default=3
            max_delay (int): maximum amount of time to wait between retries.
                default=60
        """
        return cls._backoff(_full_jitter_backoff(
            retries=retries, delay=delay, max_delay=max_delay), catch_extra_error_codes)

    @classmethod
    def backoff(cls, tries=10, delay=3, backoff=1.1, catch_extra_error_codes=None):
        """
        Retry calling the Cloud decorated function using an exponential backoff.

        Compatibility for the original implementation of CloudRetry.backoff that
        did not provide configurable backoff strategies. Developers should use
        CloudRetry.exponential_backoff instead.

        Kwargs:
            tries (int): Number of times to try (not retry) before giving up
                default=10
            delay (int or float): Initial delay between retries in seconds
                default=3
            backoff (int or float): backoff multiplier e.g. value of 2 will
                double the delay each retry
                default=1.1
        """
        return cls.exponential_backoff(
            retries=tries - 1, delay=delay, backoff=backoff, max_delay=None, catch_extra_error_codes=catch_extra_error_codes)



from __future__ import absolute_import, division, print_function


__metaclass__ = type

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _tmplt_area_range(arange):
    command = "area {area} range {range}".format(**arange)
    if arange.get("not_advertise") is True:
        command += " not-advertise"
    if "cost" in arange:
        command += " cost {cost}".format(**arange)
    return command


def _tmplt_default_information(proc):
    command = "default-information originate"
    if "always" in proc["default_information"] and proc["default_information"]["always"]:
        command += " always"
    if "route_map" in proc["default_information"]:
        command += " route-map"
        command += " {default_information[route_map]}".format(**proc)
    return command


def _tmplt_log_adjacency_changes(proc):
    command = "log-adjacency-changes"
    if proc.get("log_adjacency_changes").get("detail", False) is True:
        command += " detail"
    return command


def _tmplt_area_authentication(area):
    command = "area {area_id} authentication".format(**area)
    if area.get("authentication", {}).get("message_digest"):
        command += " message-digest"
    return command


def _tmplt_max_lsa(proc):
    max_lsa = proc["max_lsa"]
    command = "max-lsa {max_non_self_generated_lsa}".format(**max_lsa)
    if max_lsa.get("threshold"):
        command += " {threshold}".format(**max_lsa)
    if max_lsa.get("warning_only"):
        command += " warning-only"
    if max_lsa.get("ignore_time"):
        command += " ignore-time {ignore_time}".format(**max_lsa)
    if max_lsa.get("ignore_count"):
        command += " ignore-count {ignore_count}".format(**max_lsa)
    if max_lsa.get("reset_time"):
        command += " reset-time {reset_time}".format(**max_lsa)
    return command


def _tmplt_default_information(proc):
    default_information = proc["default_information"]["originate"]
    command = "default-information originate"

    if default_information.get("set") is False:
        command = "no {0}".format(command)
    else:
        if default_information.get("always"):
            command += " always"
        if default_information.get("route_map"):
            command += " route-map {route_map}".format(**default_information)

    return command


def _tmplt_table_map(proc):
    table_map = proc["table_map"]
    command = "table-map"

    if table_map.get("name"):
        command += " {name}".format(**table_map)
    if table_map.get("filter"):
        command += " filter"

    return command


def _tmplt_max_metric(proc):
    max_metric = proc["max_metric"]
    command = "max-metric router-lsa"

    if max_metric.get("router_lsa", {}).get("set") is False:
        command = "no {0}".format(command)
    else:
        external_lsa = max_metric.get("router_lsa", {}).get("external_lsa", {})
        include_stub = max_metric.get("router_lsa", {}).get("include_stub", {})
        on_startup = max_metric.get("router_lsa", {}).get("on_startup", {})
        summary_lsa = max_metric.get("router_lsa", {}).get("summary_lsa", {})
        if external_lsa:
            command += " external-lsa"
            if external_lsa.get("max_metric_value"):
                command += " {max_metric_value}".format(**external_lsa)
        if include_stub:
            command += " include-stub"
        if on_startup:
            command += " on-startup"
            if on_startup.get("wait_period"):
                command += " {wait_period}".format(**on_startup)
            if on_startup.get("wait_for_bgp_asn"):
                command += " wait-for bgp {wait_for_bgp_asn}".format(**on_startup)
        if summary_lsa:
            command += " summary-lsa"
            if summary_lsa.get("max_metric_value"):
                command += " {max_metric_value}".format(**summary_lsa)

    return command


def _tmplt_area_nssa(area):
    nssa = area["nssa"]
    command = "area {area_id} nssa".format(**area)
    if nssa.get("set") is False:
        command = "no {0}".format(command)
    else:
        for attrib in [
            "no_summary",
            "no_redistribution",
            "default_information_originate",
        ]:
            if nssa.get(attrib):
                command += " {0}".format(attrib.replace("_", "-"))
    return command


def _tmplt_area_nssa_translate(area):
    translate = area["nssa"]["translate"]["type7"]
    command = "area {area_id} nssa translate type7".format(**area)
    for attrib in ["always", "never", "supress_fa"]:
        if translate.get(attrib):
            command += " {0}".format(attrib.replace("_", "-"))
    return command


def _tmplt_area_ranges(arange):
    command = "area {area_id} range {prefix}".format(**arange)
    if arange.get("not_advertise") is True:
        command += " not-advertise"
    if "cost" in arange:
        command += " cost {cost}".format(**arange)
    return command


def _tmplt_area_ranges(arange):
    command = "area {area_id} range {prefix}".format(**arange)
    if arange.get("not_advertise") is True:
        command += " not-advertise"
    if "cost" in arange:
        command += " cost {cost}".format(**arange)
    return command


def _tmplt_summary_address(proc):
    command = "summary-address {prefix}".format(**proc)
    if proc.get("tag"):
        command += " tag {tag}".format(**proc)
    elif proc.get("not_advertise"):
        command += " not-advertise"
    return command


def _tmplt_area_stub(area):
    stub = area["stub"]
    command = "area {area_id} stub".format(**area)
    if stub.get("set") is False:
        command = "no {0}".format(command)
    elif stub.get("no_summary"):
        command += " no-summary"
    return command


def _tmplt_redistribute(redis):
    command = "redistribute {protocol}".format(**redis)
    if redis.get("id"):
        command += " {id}".format(**redis)
    if redis.get("route_map"):
        command += " route-map {route_map}".format(**redis)
    return command


def _tmplt_capability_vrf_lite(proc):
    command = "capability vrf-lite"
    vrf_lite = proc["capability"]["vrf_lite"]
    if vrf_lite.get("set") is False:
        command = "no {0}".format(command)
    else:
        if vrf_lite.get("evpn"):
            command += " evpn"
    return command


class Ospfv2Template(NetworkTemplate):
    def __init__(self, lines=None):
        super(Ospfv2Template, self).__init__(lines=lines, tmplt=self)

    # fmt: off
    PARSERS = [
        {
            "name": "vrf",
            "getval": re.compile(
                r"""
                \s+vrf
                \s(?P<vrf>\S+)$""",
                re.VERBOSE,
            ),
            "setval": "vrf {{ vrf }}",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "vrf": "{{ vrf }}",
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "bfd",
            "getval": re.compile(
                r"""
                \s+(?P<bfd>bfd)$""",
                re.VERBOSE,
            ),
            "setval": "bfd",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "bfd": "{{ not not bfd }}",
                    },
                },
            },
        },
        {
            "name": "process_id",
            "getval": re.compile(
                r"""
                ospf(?:v3)*\s
                (?P<process_id>\S+)""",
                re.VERBOSE,
            ),
            "setval": "router ospf {{ process_id }}",
            "result": {
                "process_id": "{{ process_id }}",
            },
            "shared": True,
        },
        {
            "name": "down_bit_ignore",
            "getval": re.compile(
                r"""
                \s+(?P<down_bit_ignore>down-bit-ignore)$""",
                re.VERBOSE,
            ),
            "setval": "down-bit-ignore",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "down_bit_ignore": "{{ not not down_bit_ignore }}",
                    },
                },
            },
        },
        {
            "name": "capability.vrf_lite",
            "getval": re.compile(
                r"""
                \s+capability
                \s(?P<vrf_lite>vrf-lite)
                \s*(?P<evpn>evpn)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_capability_vrf_lite,
            "remval": "capability vrf-lite",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "capability": {
                            "vrf_lite": {
                                "set": "{{ True if vrf_lite is defined and evpn is undefined else None }}",
                                "evpn": "{{ not not evpn }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "auto_cost",
            "getval": re.compile(
                r"""
                \s+auto-cost\sreference-bandwidth\s
                (?P<acrb>\d+)\s(?P<unit>\S+)$""",
                re.VERBOSE,
            ),
            "setval": (
                "auto-cost reference-bandwidth"
                " {{ auto_cost.reference_bandwidth }}"
                " {{ auto_cost.unit }}"
            ),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "auto_cost": {
                            "reference_bandwidth": "{{ acrb }}",
                            "unit": "{{ unit }}",
                        },
                    },
                },
            },
        },
        {
            "name": "flush_routes",
            "getval": re.compile(
                r"""
                \s+(?P<flush_routes>flush-routes)$""",
                re.VERBOSE,
            ),
            "setval": "flush-routes",
            "result": {
                "flush_routes": "{{ not not flush_routes }}",
            },
        },
        {
            "name": "graceful_restart.set",
            "getval": re.compile(
                r"""
                \s+(?P<graceful_restart>no\sgraceful-restart)
                $""",
                re.VERBOSE,
            ),
            "setval": "graceful-restart",
            "remval": "no graceful-restart",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "graceful_restart": {
                            "set": "{{ not graceful_restart }}",
                        },
                    },
                },
            },
        },
        {
            "name": "graceful_restart.helper_disable",
            "getval": re.compile(
                r"""
                \s+graceful-restart
                \s+(?P<helper_disable>helper-disable)
                $""",
                re.VERBOSE,
            ),
            "setval": "graceful-restart helper-disable",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "graceful_restart": {
                            "helper_disable": "{{ not not helper_disable }}",
                        },
                    },
                },
            },
        },
        {
            "name": "graceful_restart.grace_period",
            "getval": re.compile(
                r"""
                \s+graceful-restart
                \s+grace-period
                \s+(?P<grace_period>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "graceful-restart helper-disable",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "graceful_restart": {
                            "grace_period": "{{ grace_period }}",
                        },
                    },
                },
            },
        },
        {
            "name": "isolate",
            "getval": re.compile(
                r"""
                \s+(?P<isolate>isolate)$""",
                re.VERBOSE,
            ),
            "setval": "isolate",
            "result": {"isolate": "{{ not not isolate }}"},
        },
        {
            "name": "log_adjacency_changes",
            "getval": re.compile(
                r"""
                \s+(?P<log>log-adjacency-changes)
                \s*(?P<detail>detail)*$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_log_adjacency_changes,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "log_adjacency_changes": {
                            "log": "{{ True if log is defined and detail is undefined else None }}",
                            "detail": "{{ True if detail is defined else None }}",
                        },
                    },
                },
            },
        },
        {
            "name": "max_lsa",
            "getval": re.compile(
                r"""
                \s+max-lsa
                \s(?P<max_gen_lsa>\d+)
                \s*(?P<threshold>\d*)
                \s*(?P<warning_only>warning-only)*
                \s*(ignore-time)*\s*(?P<ig_time>\d*)
                \s*(ignore-count)*\s*(?P<ig_count>\d*)
                \s*(reset-time)*\s*(?P<rst_time>\d*)
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_max_lsa,
            "remval": "max-lsa {{ max_lsa.max_non_self_generated_lsa }}",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "max_lsa": {
                            "max_non_self_generated_lsa": "{{ max_gen_lsa }}",
                            "threshold": "{{ threshold }}",
                            "ignore_time": "{{ ig_time }}",
                            "ignore_count": "{{ ig_count }}",
                            "reset_time": "{{ rst_time }}",
                            "warning_only": "{{ not not warning_only }}",
                        },
                    },
                },
            },
        },
        {
            "name": "mpls.traffic_eng.areas",
            "getval": re.compile(
                r"""
                \s+mpls\straffic-eng\sarea
                \s(?P<area_id>\S+)$""",
                re.VERBOSE,
            ),
            "setval": ("mpls traffic-eng area {{ area_id }}"),
            "result": {
                "mpls": {
                    "traffic_eng": {
                        "areas": [
                            {
                                "area_id": "{{ area_id }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "mpls.traffic_eng.router_id",
            "getval": re.compile(
                r"""
                \s+mpls\straffic-eng\srouter-id
                \s(?P<router_id>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": (
                "mpls traffic-eng router-id" " {{ mpls.traffic_eng.router_id }}"
            ),
            "result": {"mpls": {"traffic_eng": {"router_id": "{{ router_id }}"}}},
        },
        {
            "name": "mpls.traffic_eng.multicast_intact",
            "getval": re.compile(
                r"""
                \s+mpls\straffic-eng
                \s(?P<multicast_intact>multicast-intact)
                $""",
                re.VERBOSE,
            ),
            "setval": ("mpls traffic-eng multicast-intact"),
            "result": {
                "mpls": {
                    "traffic_eng": {
                        "multicast_intact": "{{ not not multicast_intact }}",
                    },
                },
            },
        },
        {
            "name": "name_lookup",
            "getval": re.compile(
                r"""
                \s+(?P<name_lookup>name-lookup)
                $""",
                re.VERBOSE,
            ),
            "setval": ("name-lookup"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "name_lookup": "{{ not not name_lookup }}",
                    },
                },
            },
        },
        {
            "name": "passive_interface.default",
            "getval": re.compile(
                r"""
                \s+passive-interface
                \s+(?P<default>default)
                $""",
                re.VERBOSE,
            ),
            "setval": ("passive-interface default"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "passive_interface": {"default": "{{ not not default }}"},
                    },
                },
            },
        },
        {
            "name": "rfc1583compatibility",
            "getval": re.compile(
                r"""
                \s+(?P<rfc>rfc1583compatibility)$""",
                re.VERBOSE,
            ),
            "setval": ("rfc1583compatibility"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "rfc1583compatibility": "{{ not not rfc }}",
                    },
                },
            },
        },
        {
            "name": "router_id",
            "getval": re.compile(
                r"""
                \s+router-id
                \s(?P<router_id>\S+)$""",
                re.VERBOSE,
            ),
            "setval": ("router-id" " {{ router_id }}"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "router_id": "{{ router_id }}",
                    },
                },
            },
        },
        {
            "name": "shutdown",
            "getval": re.compile(
                r"""
                \s+(?P<shutdown>shutdown)$""",
                re.VERBOSE,
            ),
            "setval": ("shutdown"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "shutdown": "{{ not not shutdown }}",
                    },
                },
            },
        },
        {
            "name": "default_information.originate",
            "getval": re.compile(
                r"""
                \s+default-information
                \s(?P<originate>originate)
                \s*(?P<always>always)*
                \s*(route-map)*
                \s*(?P<route_map>\S+)*$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_default_information,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "default_information": {
                            "originate": {
                                "set": "{{ True if originate is defined and always is undefined and route_map is undefined else None }}",
                                "always": "{{ not not always }}",
                                "route_map": "{{ route_map }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "default_metric",
            "getval": re.compile(
                r"""
                \s+default-metric
                \s(?P<default_metric>\d+)$""",
                re.VERBOSE,
            ),
            "setval": ("default-metric {{ default_metric }}"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "default_metric": "{{ default_metric }}",
                    },
                },
            },
        },
        {
            "name": "distance",
            "getval": re.compile(
                r"""
                \s+distance
                \s(?P<distance>\d+)$""",
                re.VERBOSE,
            ),
            "setval": ("distance {{ distance }}"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "distance": "{{ distance }}",
                    },
                },
            },
        },
        {
            "name": "table_map",
            "getval": re.compile(
                r"""
                \s+table-map
                \s(?P<rmap>\S+)
                \s*(?P<filter>filter)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_table_map,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "table_map": {
                            "name": "{{ rmap }}",
                            "filter": "{{ not not filter }}",
                        },
                    },
                },
            },
        },
        {
            "name": "timers.lsa_arrival",
            "getval": re.compile(
                r"""
                \s+timers
                \slsa-arrival
                \s(?P<lsa_arrival_val>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": ("timers lsa-arrival {{ timers.lsa_arrival }}"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "timers": {
                            "lsa_arrival": "{{ lsa_arrival_val }}",
                        },
                    },
                },
            },
        },
        {
            "name": "timers.lsa_group_pacing",
            "getval": re.compile(
                r"""
                \s+timers
                \slsa-group-pacing
                \s(?P<lsa_group_pacing>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "timers lsa-group-pacing {{ timers.lsa_group_pacing }}",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "timers": {
                            "lsa_group_pacing": "{{ lsa_group_pacing }}",
                        },
                    },
                },
            },
        },
        {
            "name": "timers.throttle.lsa",
            "getval": re.compile(
                r"""
                \s+timers\sthrottle\slsa
                \s(?P<start>\d+)
                \s(?P<hold>\d+)
                \s(?P<max>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "timers throttle lsa {{ timers.throttle.lsa.start_interval }}"
            " {{ timers.throttle.lsa.hold_interval }}"
            " {{ timers.throttle.lsa.max_interval }}",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "timers": {
                            "throttle": {
                                "lsa": {
                                    "start_interval": "{{ start }}",
                                    "hold_interval": "{{ hold }}",
                                    "max_interval": "{{ max }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "timers.throttle.spf",
            "getval": re.compile(
                r"""
                \s+timers\sthrottle\sspf
                \s(?P<initial>\d+)
                \s(?P<min>\d+)
                \s(?P<max>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "timers throttle spf {{ timers.throttle.spf.initial_spf_delay }}"
            " {{ timers.throttle.spf.min_hold_time }}"
            " {{ timers.throttle.spf.max_wait_time }}",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "timers": {
                            "throttle": {
                                "spf": {
                                    "initial_spf_delay": "{{ initial }}",
                                    "min_hold_time": "{{ min }}",
                                    "max_wait_time": "{{ max }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "area.default_cost",
            "getval": re.compile(
                r"""
                \s+area\s(?P<area_id>\S+)\s
                default-cost\s(?P<default_cost>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "area {{ area_id }} default-cost {{ default_cost }}",
            "compval": "default_cost",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "areas": {
                            "{{ area_id }}": {
                                "area_id": "{{ area_id }}",
                                "default_cost": "{{ default_cost|int }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "area.authentication",
            "getval": re.compile(
                r"""
                \s+area\s(?P<area_id>\S+)
                \s(?P<auth>authentication)
                \s*(?P<md>message-digest)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_area_authentication,
            "remval": "area {{ area_id }} authentication",
            "compval": "authentication",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "areas": {
                            "{{ area_id }}": {
                                "area_id": "{{ area_id }}",
                                "authentication": {
                                    "set": "{{ True if auth is defined and md is undefined }}",
                                    "message_digest": "{{ True if md is defined else False }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "area.filter_list",
            "getval": re.compile(
                r"""
                \s+area\s(?P<area_id>\S+)
                \sfilter-list
                \sroute-map\s(?P<rmap>\S+)
                \s(?P<dir>\S+)$""",
                re.VERBOSE,
            ),
            "setval": "area {{ area_id }} filter-list route-map {{ route_map }} {{ direction }}",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "areas": {
                            "{{ area_id }}": {
                                "area_id": "{{ area_id }}",
                                "filter_list": [
                                    {
                                        "route_map": "{{ rmap }}",
                                        "direction": "{{ dir }}",
                                    },
                                ],
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "redistribute",
            "getval": re.compile(
                r"""
                \s+redistribute
                \s(?P<protocol>\S+)
                \s*(?P<id>\S+)*
                \sroute-map\s(?P<rmap>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_redistribute,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "redistribute": [
                            {
                                "protocol": "{{ protocol }}",
                                "id": "{{ id }}",
                                "route_map": "{{ rmap }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "area.nssa",
            "getval": re.compile(
                r"""
                \s+area\s(?P<area_id>\S+)
                \s(?P<nssa>nssa)
                \s*(?P<no_sum>no-summary)*
                \s*(?P<no_redis>no-redistribution)*
                \s*(?P<def_info>default-information-originate)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_area_nssa,
            "remval": "area {{ area_id }} nssa",
            "compval": "nssa",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "areas": {
                            "{{ area_id }}": {
                                "area_id": "{{ area_id }}",
                                "nssa": {
                                    "set": "{{ True if nssa is defined and no_sum is undefined and no_redis is undefined and def_info is undefined }}",
                                    "no_summary": "{{ not not no_sum }}",
                                    "no_redistribution": "{{ not not no_redis }}",
                                    "default_information_originate": "{{ not not def_info }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "area.nssa.translate",
            "getval": re.compile(
                r"""
                \s+area\s(?P<area_id>\S+)\snssa
                \stranslate
                \stype7
                \s(?P<choice>always|never)
                \s*(?P<supress_fa>supress-fa)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_area_nssa_translate,
            "compval": "nssa.translate",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "areas": {
                            "{{ area_id }}": {
                                "area_id": "{{ area_id }}",
                                "nssa": {
                                    "translate": {
                                        "type7": {
                                            "always": '{{ True if choice == "always" else None }}',
                                            "never": '{{ True if choice == "never" else None }}',
                                            "supress_fa": "{{ not not supress_fa }}",
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "area.ranges",
            "getval": re.compile(
                r"""
                \s+area\s(?P<area_id>\S+)
                \srange\s(?P<prefix>\S+)
                \s*(cost)*\s*(?P<cost>\d+)*
                \s*(?P<not_adver>not-advertise)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_area_ranges,
            "remval": "area {{ area_id }} range {{ prefix }}",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "areas": {
                            "{{ area_id }}": {
                                "area_id": "{{ area_id }}",
                                "ranges": [
                                    {
                                        "prefix": "{{ prefix }}",
                                        "cost": "{{ cost }}",
                                        "not_advertise": "{{ not not not_adver }}",
                                    },
                                ],
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "summary_address",
            "getval": re.compile(
                r"""
                \s+summary-address
                \s(?P<prefix>\S+)
                \s*(?P<not_adver>not-advertise)*
                \s*(tag)*\s*(?P<tag>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_summary_address,
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "summary_address": [
                            {
                                "prefix": "{{ prefix }}",
                                "not_advertise": "{{ not not not_adver }}",
                                "tag": "{{ tag }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "area.stub",
            "getval": re.compile(
                r"""
                \s+area\s(?P<area_id>\S+)
                \s(?P<stub>stub)
                \s*(?P<no_summary>no-summary)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_area_stub,
            "remval": "area {{ area_id }} stub",
            "compval": "stub",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "areas": {
                            "{{ area_id }}": {
                                "area_id": "{{ area_id }}",
                                "stub": {
                                    "set": "{{ True if stub is defined and no_summary is undefined else None }}",
                                    "no_summary": "{{ not not no_summary }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "maximum_paths",
            "getval": re.compile(
                r"""
                \s+maximum-paths
                \s(?P<maximum_paths>\d+)$""",
                re.VERBOSE,
            ),
            "setval": ("maximum-paths {{ maximum_paths }}"),
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {"maximum_paths": "{{ maximum_paths }}"},
                },
            },
        },
        {
            "name": "max_metric",
            "getval": re.compile(
                r"""
                \s+max-metric
                \s+(?P<router_lsa>router-lsa)
                \s*(?P<external_lsa>external-lsa)*
                \s*(?P<max_metric_value>\d+)*
                \s*(?P<include_stub>include-stub)*
                \s*(?P<on_startup>on-startup)*
                \s*(?P<wait_period>\d+)*
                \s*(wait-for\sbgp)*
                \s*(?P<bgp_asn>\d+)*
                \s*(?P<summary_lsa>summary-lsa)*
                \s*(?P<sum_lsa_max_metric_value>\d+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_max_metric,
            "remval": "max-metric router-lsa",
            "result": {
                "vrfs": {
                    '{{ "vrf_" + vrf|d() }}': {
                        "max_metric": {
                            "router_lsa": {
                                "set": "{{ True if router_lsa is defined and external_lsa is undefined else None }}",
                                "external_lsa": {
                                    "set": "{{ True if external_lsa is defined and max_metric_value is undefined else None }}",
                                    "max_metric_value": "{{ max_metric_value }}",
                                },
                                "include_stub": "{{ not not include_stub }}",
                                "on_startup": {
                                    "set": "{{ True if on_startup is defined and (wait_period and bgp_asn) is undefined else None }}",
                                    "wait_period": "{{ wait_period }}",
                                    "wait_for_bgp_asn": "{{ bgp_asn }}",
                                },
                                "summary_lsa": {
                                    "set": "{{ True if summary_lsa is defined and sum_lsa_max_metric_value is undefined else None }}",
                                    "max_metric_value": "{{ sum_lsa_max_metric_value }}",
                                },
                            },
                        },
                    },
                },
            },
        },
    ]
    # fmt: on
